import inspect

from posthog.client import require


class ToolParser:

    def __init__(self):
        pass

    def _get_doc(self, func):
        doc = inspect.getdoc(func)
        if "Args:" in doc:
            doc = doc.split("Args:")[0]
        doc = doc.replace("\n","").strip()
        return doc

    def _get_params(self, func):
        
        types_map = {
            "str":"string"
        }

        sig = inspect.signature(func)
        params = []

        for name, param in sig.parameters.items():
            # Gestione sicura del tipo
            param_type = "string"  # default
            if param.annotation != inspect.Parameter.empty:
                type_name = param.annotation.__name__
                param_type = types_map.get(type_name, "string")
            
            p = {
                    "name":name, 
                    "type": param_type, 
                }
            
            default = param.default

            if default != inspect.Parameter.empty:
                p["default"] = default
                p["required"] = False
            else:
                p["required"] = True

            params.append(p)
        
        if len(params)==0:
            return params

        doc = inspect.getdoc(func)
        doc = doc.split("Args:")[1]

        for param in params:
            delimiter = param["name"]+" ("+param["type"]+")"+":"
            desc = doc[doc.find(delimiter)+len(delimiter):]
            desc = desc.split("\n")[0].strip()
            param["description"] = desc

        return params

    def parse(self, func):

        tool = {
                "type": "function",
                "function": {
                    "name": func.__name__,
                    "description": self._get_doc(func),
                }
        }

        params = self._get_params(func)

        if len(params)==0:
            return tool

        tool["function"]["parameters"] = {"type":"object", "properties":{}}
        required = []
        for param in params:
            tool["function"]["parameters"]["properties"][param["name"]] = {
                "type":param["type"],
                "description":param["description"]
            }
            if param["required"]:
                required.append(param["name"])

        if len(required)>0:
            tool["function"]["parameters"]["required"] = required

        return tool

    


from typing import List
from .prompt import Prompt, _PromptParser

class PromptChain(Prompt):

    """
    PromptChain class for handling sequences of prompts, the output of one prompt is used as context for the next one.

    .prompt file example:
    --------
    ```
    <promptchain>
        <prompt>
            What is the capital of {country}?
        </prompt>
        <prompt>
            What is its population?
        </prompt>
    </promptchain>
    ```

    Examples
    --------
    Create a chain from a sequence of prompts:
    ```
    prompts = [
        Prompt(prompt="What is the capital of France?"),
        Prompt(prompt="What is its population?")
    ]
    chain = PromptChain(prompts=prompts)
    ```

    Create a chain from a file:
    ```
    chain = PromptChain(
        promptchain_id="city_analysis",
        prompts_data=[{"country": "France"}]
    )
    ```
    """

    def __init__(self, 
                 promptchain_id: str = None,
                 prompts_data: list[dict] = None,
                 prompts: List[Prompt] = None,
                 response_type: type | None = None):

        """
        Initialize a new PromptChain instance.

        Parameters
        ----------
        promptchain_id : str, optional
            A .prompt file name for loading a prompt chain from file
        prompts_data : list[dict], optional
            List of dictionaries containing formatting data for each prompt
        prompts : List[Prompt], optional
            Direct list of Prompt objects to form the chain if promptchain_id is not provided

        Raises
        ------
        - ValueError
            If neither promptchain_id nor prompts is provided
        """
        if promptchain_id is not None:
            self._prompts, self._response_type = _PromptChainParser().parse(promptchain_id)
            for i in range(len(self._prompts)):
                self._prompts[i] = Prompt(prompt=self._prompts[i], prompt_data=prompts_data[i])
        elif prompts is not None:
            self._prompts = prompts
        else:
            raise ValueError("Either promptchain_id or prompts must be provided")
        self._size = len(self._prompts)
        self.response_type = response_type

    def _format(self, index: int, context: str | None = None) -> str:
        """
        Format a specific prompt in the chain with optional context.

        This method formats the prompt at the specified index, optionally including
        context from previous prompts' responses.

        Parameters
        ----------
        index : int
            Index of the prompt to format
        context : str, optional
            Context from previous prompts' responses to include

        Returns
        -------
        str
            The formatted prompt text with optional context
        """
        if context is None:
            return str(self._prompts[index])
        else:
            return str(self._prompts[index]) + "\n\n" + context

    def __str__(self) -> str:
        """
        Get the string representation of the prompt chain.

        Returns
        -------
        str
            All prompts in the chain joined by newlines
        """
        return "\n".join([str(prompt) for prompt in self._prompts])

    def __repr__(self) -> str:
        """
        Get the official string representation of the prompt chain.

        Returns
        -------
        str
            All prompts in the chain joined by newlines
        """
        return self.__str__()


class _PromptChainParser(_PromptParser):
    def _parse(self, prompt_dict):
        prompt_dict = prompt_dict["promptchain"]
        return prompt_dict["prompt"], prompt_dict.get("@response_type")
        

if __name__ == "__main__":
    parser = _PromptChainParser()
    print(parser.parse("test_chain"))

    


from ._base_model import BaseModel
from .model import Model
from ._prompt_executor import PromptExecutorMixin
from ._response_processor import ResponseProcessorMixin
from typing import List, Dict, Union
import asyncio
from ..prompts.prompt_chain import PromptChain
from ..prompts.prompt import Prompt

class MultiModel(BaseModel, PromptExecutorMixin, ResponseProcessorMixin):
    """
    A class to execute prompts across multiple AI models in parallel.
    
    MultiModel manages a collection of AI models and enables parallel execution of prompts
    across all models. It's particularly useful for comparing model responses or
    implementing ensemble approaches.

    Examples
    --------
    Basic comparison of models:
    ```
    models = [
        {"provider": "openai", "model": "gpt-4"},
        {"provider": "anthropic", "model": "claude-3"}
    ]
    multi_model = MultiModel(models=models)
    prompt = Prompt(
        prompt="What is 2+2?",
        response_type=int
    )
    responses = multi_model.ask(prompt)
    for resp in responses:
        print(f"{resp['model']['name']}: {resp['response']}")
    ```
    """

    def __init__(
        self, 
        models: List[Dict[str, str]], 
        count_tokens: bool = False, 
        count_cost: bool = False
    ):
        """
        Initialize a new MultiModel instance.

        Parameters
        ----------
        models : List[Dict[str, str]]
            List of dictionaries with provider and model information
        count_tokens : bool, optional
            Whether to count tokens for each request
        count_cost : bool, optional
            Whether to calculate costs for each request
        """
        super().__init__(count_tokens, count_cost)
        self._models = [
            Model(
                model['provider'],
                model['model'],
                count_tokens=count_tokens,
                count_cost=count_cost
            ) for model in models
        ]

    async def _task(self, model: Model, prompt: Union[str, Prompt, PromptChain]) -> Dict:
        """
        Execute a single model task asynchronously.

        Parameters
        ----------
        model : Model
            The model instance to use
        prompt : Union[str, Prompt, PromptChain]
            The prompt to process

        Returns
        -------
        Dict
            Dictionary containing:
            - response: The model's response
            - prompt: The original prompt
            - model: Dictionary with provider and model name
            - tokens: Token counts (if enabled)
            - cost: Cost calculation (if enabled)
        """
        response = await self._execute_async(prompt, model._agent)
        return self._process_response(
            prompt,
            response,
            model.provider,
            model.model,
            self._count_tokens,
            self._count_cost
        )

    async def _ask_async(self, prompt: Union[str, Prompt, PromptChain]) -> List[Dict]:
        """
        Ask all models asynchronously.

        Parameters
        ----------
        prompt : Union[str, Prompt, PromptChain]
            The prompt to process across all models

        Returns
        -------
        List[Dict]
            List of response dictionaries, one per model, each containing:
            - response: The model's response
            - prompt: The original prompt
            - model: Dictionary with provider and model name
            - tokens: Token counts (if enabled)
            - cost: Cost calculation (if enabled)

        Examples
        --------
        Using async/await:
            >>> responses = await multi_model.ask_async("What is 2+2?")
            >>> for resp in responses:
            ...     print(f"{resp['model']['name']}: {resp['response']}")
        """
        tasks = [self._task(model, prompt) for model in self._models]
        return await asyncio.gather(*tasks)

    def ask(self, prompt: Union[str, Prompt]) -> List[Dict]:
        """
        Ask all models.

        Parameters
        ----------
        prompt : Union[str, Prompt]
            The prompt to process across all models

        Returns
        -------
        List[Dict]
            List of response dictionaries, one per model, each containing:
            - response: The model's response
            - prompt: The original prompt
            - model: Dictionary with provider and model name
            - tokens: Token counts (if enabled)
            - cost: Cost calculation (if enabled)

        """
        return asyncio.run(self.ask_async(prompt))


if __name__ == "__main__":
    # Initialize with counting preferences
    multi_model = MultiModel(
        models=[
            {'provider': 'openai', 'model': 'gpt-4o-mini'},
            {'provider': 'deepseek', 'model': 'deepseek-chat'},
        ],
        count_tokens=True,
        count_cost=True
    )
    
    # Example with simple prompt
    results = multi_model.ask("What is the capital of France?")
    print("\nSimple prompt results:")
    for i, result in enumerate(results, 1):
        print(f"\nModel {i} Response:")
        print(result["output"])
        if "tokens" in result:
            print("Token Count:")
            print(f"Input tokens: {result['tokens']['input_tokens']:,}")
            print(f"Output tokens: {result['tokens']['output_tokens']:,}")
            print(f"Total tokens: {result['tokens']['total_tokens']:,}")
        if "cost" in result:
            print("Cost:")
            print(f"Input cost: ${result['cost']['input_cost']:.6f}")
            print(f"Output cost: ${result['cost']['output_cost']:.6f}")
            print(f"Total cost: ${result['cost']['total_cost']:.6f}")
    
    # Example with prompt chain
    chain = PromptChain([
        "What is the capital of France?",
        "Based on the previous answer, what is its population?"
    ])
    results = multi_model.ask(chain)
    print("\nPrompt chain results:")
    for i, result in enumerate(results, 1):
        print(f"\nModel {i} Response:")
        print(result["output"])
        if "tokens" in result:
            print("Token Count:")
            print(f"Input tokens: {result['tokens']['input_tokens']:,}")
            print(f"Output tokens: {result['tokens']['output_tokens']:,}")
            print(f"Total tokens: {result['tokens']['total_tokens']:,}")
        if "cost" in result:
            print("Cost:")
            print(f"Input cost: ${result['cost']['input_cost']:.6f}")
            print(f"Output cost: ${result['cost']['output_cost']:.6f}")
            print(f"Total cost: ${result['cost']['total_cost']:.6f}")



from whois import whois

def domain_whois(domain:str):

    """
    Get the whois for a given domain:
    
    Args:
        domain (string): the domain
    """

    result = whois(domain)
    return str(result)


from os.path import join
from typing import List
from monoai.conf import Conf
import xmltodict


class Prompt:
    """
    A class to handle text prompts with optional response type specification.
    
    The Prompt class can be initialized either with a direct prompt string or by loading
    a prompt from a .prompt file. It supports response type specification and
    data formatting through Python's string formatting.

    .prompt format
    --------
    Simple prompt
    ```
    <prompt>
        Hello, coicoi!
    </prompt>
    ```
    In this case you could also use plain text without the prompt tag:
    ```
    Hello, coicoi!
    ```
    Prompt with data formatting
    ```
    <prompt>
        Hello, {name}!
    </prompt>
    ```
    Prompt with response type
    ```
    <prompt response_type="int">
        What is 2+2?
    </prompt>
    ```

    Examples
    --------
    Simple prompt with response type:
    ```
    prompt = Prompt(prompt="What is 2+2?", response_type=int)
    ```
    Prompt with data formatting:
    ```
    prompt = Prompt(
        prompt="What is the capital of {country}?",
        prompt_data={"country": "France"},
        response_type=str
    )
    ```
    Load prompt from file:
    ```
    prompt = Prompt(prompt_id="math_question")
    ```
    math_question.prompt
    ```
    <prompt response_type="int">
        What is 2+2?
    </prompt>
    ```
    """

    def __init__(self, 
                 prompt_id: str | None = None,
                 prompt_data: dict | None = None,
                 prompt: str | None = None, 
                 response_type: type | None = None):
        """
        Initialize a new Prompt instance.

        Parameters
        ----------
        prompt_id : str, optional
            A .prompt file name for loading a prompt from file
        prompt_data : dict, optional
            Dictionary of values for formatting the prompt
        prompt : str, optional
            Direct prompt text if prompt_id is not provided
        response_type : type | BaseModel, optional
            Expected type of the response or a Pydantic BaseModel for json schema response

        Raises
        ------
        ValueError
            If neither prompt_id nor prompt is provided
        """
        prompt_response_type = None
        if prompt_id is not None:
            self._prompt, prompt_response_type = _PromptParser().parse(prompt_id)
        elif prompt is not None:
            self._prompt = prompt
        else:
            raise ValueError("Either prompt_id or prompt must be provided")

        if prompt_data is not None:
            self._prompt = self._prompt.format(**prompt_data)

        if prompt_response_type is not None:
            self.response_type = prompt_response_type
        else:
            self.response_type = response_type

    def __str__(self) -> str:
        """
        Get the string representation of the prompt.

        Returns
        -------
        str
            The prompt text
        """
        return self._prompt
    
    def __repr__(self) -> str:
        """
        Get the official string representation of the prompt.

        Returns
        -------
        str
            The prompt text
        """
        return self.__str__()


    def as_dict(self)->dict:
        return {"type":"user", "content":self.__str__()}


class _PromptParser:
    def parse(self, prompt_id: str):
        
        if prompt_id.endswith(".prompt"):
            prompt_path=prompt_id
        else:
            prompt_path = join(Conf()["prompts_path"], prompt_id+".prompt")
        prompt_file = open(prompt_path)
        prompt_text = prompt_file.read()        
        if self._very_dumb_xml_check(prompt_text):
            prompt_dict = xmltodict.parse(prompt_text)
            return self._parse(prompt_dict), None
        else:
            return prompt_text, None

    def _very_dumb_xml_check(self, text: str):
        return text[0]=="<" and text[-1]==">"
    
    def _parse(self, prompt_dict: dict):
        if isinstance(prompt_dict["prompt"], dict):
            prompt_dict = prompt_dict["prompt"]
            return prompt_dict["#text"], self._type_from_str(prompt_dict["@response_type"])
        return prompt_dict["prompt"]
    
    """
    def _parse(self, text: str):
        root = ET.fromstring(text)
        if root.tag == "promptchain":
            return [prompt.text for prompt in root.findall("prompt")], root.get("response_type")
        elif root.tag == "iterativeprompt":
            prompt_memory = root.find("prompt_memory")
            return root.find("prompt").text, root.find("prompt_memory").text
        else:
            return root.text, self._type_from_str(root.get("response_type"))
    """
    
    def _type_from_str(self, type_str: str):
        
        if type_str is None:
            return None
        
        type_mapping = {
            "str": str,
            "int": int,
            "float": float,
            "bool": bool,
            "list": List[str],
            "dict": dict,
            "tuple": tuple,
            "set": set,
            "None": type(None)
        }
        
        assert type_str in type_mapping, f"Type {type_str} not found in type_mapping"
        return type_mapping[type_str]
    
    
if __name__ == "__main__":
    parser = _PromptParser()
    print(parser.parse("test"))




"""
RAG is a module that provides a high-level interface for performing semantic search queries
against a vector database. It supports multiple vector database backends and
embedding providers for flexible deployment scenarios.
"""

from .rag import RAG
from .vectordb import ChromaVectorDB
from .documents_builder import DocumentsBuilder

__all__ = ['RAG', 'ChromaVectorDB', 'DocumentsBuilder'] 


def scrape_web(url: str, engine: str = "requests", deep: bool = False):
    """Scrape a webpage using the specified engine.
    
    Args:
        url: The url to scrape
        engine: The engine to use (requests, tavily, selenium). Default is requests
        deep: If using tavily, whether to use the advanced extraction mode. Default is False
    
    Returns:
        A dictionary containing:
            html: The html of the page (not available if using tavily)
            text: The content of the page merged into a single string
            
    Raises:
        ValueError: If an invalid engine is specified
        
    Examples:
        Basic usage with requests engine:
        ```python
        result = scrape_web("https://www.scrapethissite.com")
        print(result["html"])  # print the html of the page
        print(result["text"])  # print the content of the page merged into a single string
        ```
        
        Using tavily with deep extraction:
        ```python
        result = scrape_web("https://example.com", engine="tavily", deep=True)
        print(result["text"])  # print the extracted content
        ```
    """
    
    if engine == "requests":
        scraper = _RequestsScraper()
    elif engine == "tavily":
        scraper = _TavilyScraper(deep=deep)
    elif engine == "selenium":
        scraper = _SeleniumScraper()
    else:
        raise ValueError(f"Invalid engine: {engine} (must be 'requests', 'tavily', or 'selenium')")
    
    response, text_response = scraper.scrape(url)
    return {"html": response, "text": text_response}


from bs4 import BeautifulSoup

class _BaseScraper():
    """Base class for web scrapers."""

    def __init__(self):
        """Initialize the base scraper."""
        pass

    def scrape(self, url: str):
        """Scrape a webpage.
        
        Args:
            url: The URL to scrape
            
        Returns:
            A tuple of (html_content, text_content)
        """
        pass

    def _extract_text(self, content: str, remove_whitespace: bool = True) -> str:
        """Extract text content from HTML, optionally removing extra whitespace.
        
        Args:
            content: The HTML content to extract text from
            remove_whitespace: Whether to remove extra whitespace and normalize spaces
            
        Returns:
            The extracted text content
        """
        soup = BeautifulSoup(content, 'html.parser')
        
        # Remove script and style elements
        for script in soup(["script", "style"]):
            script.decompose()
        
        # Get text
        text = soup.get_text()
        
        if remove_whitespace:
            # Break into lines and remove leading and trailing space on each
            lines = (line.strip() for line in text.splitlines())
            # Break multi-headlines into a line each
            chunks = (phrase.strip() for line in lines for phrase in line.split("  "))
            # Drop blank lines
            text = ' '.join(chunk for chunk in chunks if chunk)
        
        return text 


import requests

class _RequestsScraper(_BaseScraper):
    """Scraper that uses the requests library for basic web scraping."""

    def __init__(self):
        """Initialize the requests scraper."""
        super().__init__()

    def scrape(self, url: str):
        """Scrape a webpage using requests.
        
        Args:
            url: The URL to scrape
            
        Returns:
            A tuple of (html_content, text_content)
        """
        response = requests.get(url)
        return response.text, self._extract_text(response.text)
    

from tavily import TavilyClient
from monoai.keys.keys_manager import load_key

class _TavilyScraper(_BaseScraper):
    """Scraper that uses the Tavily API for advanced content extraction."""

    def __init__(self, deep: bool = False):
        """Initialize the Tavily scraper.
        
        Args:
            deep: Whether to use advanced extraction mode
        """
        super().__init__()
        load_key("tavily")
        self._client = TavilyClient()
        self._deep = deep

    def scrape(self, url: str):
        """Scrape a webpage using Tavily.
        
        Args:
            url: The URL to scrape
            
        Returns:
            A tuple of (None, extracted_content) since Tavily doesn't return raw HTML
        """
        response = self._client.extract(url, extract_depth="advanced" if self._deep else "basic")
        response = response["results"][0]
        return None, response["raw_content"]

from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.common.exceptions import TimeoutException
import time

class _SeleniumScraper(_BaseScraper):
    """A scraper that uses Selenium to handle dynamic content and JavaScript.
    
    This is useful for websites that require JavaScript execution to load content.
    """
    
    def __init__(self, headless: bool = True, wait_time: int = 10):
        """Initialize the Selenium scraper.
        
        Args:
            headless: Whether to run Chrome in headless mode. Default is True
            wait_time: Maximum time to wait for elements to load in seconds. Default is 10
        """
        super().__init__()
        self._wait_time = wait_time
        self._setup_driver(headless)
    
    def _setup_driver(self, headless: bool):
        """Set up the Chrome WebDriver with specified options.
        
        Args:
            headless: Whether to run Chrome in headless mode
        """
        chrome_options = Options()
        if headless:
            chrome_options.add_argument("--headless")
        chrome_options.add_argument("--no-sandbox")
        chrome_options.add_argument("--disable-dev-shm-usage")
        chrome_options.add_argument("--disable-gpu")
        chrome_options.add_argument("--window-size=1920,1080")
        
        self._driver = webdriver.Chrome(options=chrome_options)
        self._wait = WebDriverWait(self._driver, self._wait_time)
    
    def scrape(self, url: str):
        """Scrape a webpage using Selenium.
        
        Args:
            url: The URL to scrape
            
        Returns:
            A tuple of (html_content, text_content)
        """
        try:
            self._driver.get(url)
            # Wait for the page to load
            time.sleep(2)  # Basic wait for initial page load
            
            # Get the page source after JavaScript execution
            html_content = self._driver.page_source
            text_content = self._extract_text(html_content)
            
            return html_content, text_content
            
        except Exception as e:
            print(f"Error scraping {url}: {str(e)}")
            return None, None
            
    def __del__(self):
        """Clean up the WebDriver when the object is destroyed."""
        if hasattr(self, '_driver'):
            self._driver.quit()



from ..models import CollaborativeModel
from typing import List, Dict, Union, Callable
from ..tools._base_tools import BaseTool

class CollaborativeAgent(CollaborativeModel):

    def __init__(
        self,
        models: List[Dict[str, str]],
        aggregator: Dict[str, str],
        tools: List[Union[BaseTool, Callable]] = [],
        count_tokens: bool = False,
        count_cost: bool = False
    ):
        super().__init__(models, aggregator, count_tokens, count_cost)

        tools_functions = [tool.get_tool() if isinstance(tool, BaseTool) else tool for tool in tools]

        for model in self._multi_model._models:
            for tool in tools_functions:
                model._agent._register_tool(tool)

    def run(self, prompt: str):
        return super().ask(prompt)



from monoai.models import Model
from monoai.agents import Agent
from monoai.agents.agentic_loop import ProgrammaticAgenticLoop
from datetime import datetime
from monoai.tools import domain_whois, search_web


def get_current_time(location: str):

    """
    Get the current time in a given location

    Args:
        location (str): the location
    """

    return datetime.now().strftime("%H:%M")

def get_user_location():
    
    """
    Get the location of the current user
    """
    
    return "Catania"

model = Model(provider="openai", model="gpt-4.1-nano")
agentic_loop = ProgrammaticAgenticLoop(model=model, 
              debug=True,
              max_iter=10)

agent = Agent(model=model, 
              tools=[get_current_time, get_user_location], 
              paradigm=agentic_loop)

resp = agent.run("Che ora è? Fai così: 1 - ottieni la località, 2 - ottieni l'ora")
print(resp)


import chromadb
from litellm import embedding
from typing import List, Dict, Any, Optional, Union

class _BaseVectorDB:
    """
    Abstract base class for vector database implementations.
    
    This class defines the interface that all vector database backends must implement.
    It provides common functionality for document embedding and defines the contract
    for database operations like creation, addition, querying, and deletion.
    
    The base class handles:
    - Configuration of embedding models and providers
    - Document vectorization using the specified embedding model
    - Common interface for all vector database operations
    
    Attributes:
        _name (str): Name of the vector database collection
        _vectorizer_provider (str): The embedding provider (e.g., "openai", "anthropic")
        _vectorizer_model (str): The specific embedding model to use
    
    Note:
        This is an abstract base class. Use concrete implementations like
        ChromaVectorDB for actual vector database operations.
    """

    def __init__(self, name: Optional[str] = None, 
                 vectorizer_provider: Optional[str] = None, 
                 vectorizer_model: Optional[str] = None):
        """
        Initialize the base vector database.
        
        Parameters:
        -----------
        name : str, optional
            Name of the vector database collection. If None, no collection
            is created during initialization.
            
        vectorizer_provider : str, optional
            The embedding provider to use for vectorization.
            Examples: "openai", "anthropic", "cohere"
            
        vectorizer_model : str, optional
            The specific embedding model to use.
            Examples: "text-embedding-ada-002", "text-embedding-3-small"
        """
        self._name = name
        self._vectorizer_provider = vectorizer_provider
        self._vectorizer_model = vectorizer_model

    def _embed(self, documents: List[str]) -> List[List[float]]:
        """
        Convert text documents into vector embeddings.
        
        This method uses the configured embedding model to convert text documents
        into numerical vector representations that can be stored and queried
        in the vector database.
        
        Parameters:
        -----------
        documents : List[str]
            List of text documents to convert into embeddings.
            
        Returns:
        --------
        List[List[float]]
            List of vector embeddings, where each embedding is a list of floats.
            
        Examples:
        ---------
        ```python
        # Convert documents to embeddings
        docs = ["Hello world", "Machine learning is fascinating"]
        embeddings = vector_db._embed(docs)
        print(f"Generated {len(embeddings)} embeddings")
        print(f"Each embedding has {len(embeddings[0])} dimensions")
        ```
        """
        result = embedding(
            model=self._vectorizer_model,
            input=documents
        )
        return result

    def new(self, name: str) -> None:
        """
        Create a new vector database collection.
        
        This method should be implemented by concrete subclasses to create
        a new collection with the specified name.
        
        Parameters:
        -----------
        name : str
            Name of the new collection to create.
            
        Note:
        -----
        This is an abstract method that must be implemented by subclasses.
        """
        pass

    def add(self, documents: List[str], metadatas: Optional[List[Dict]] = None, 
            ids: Optional[List[str]] = None) -> None:
        """
        Add documents to the vector database.
        
        This method should be implemented by concrete subclasses to add
        documents along with their metadata and IDs to the vector database.
        
        Parameters:
        -----------
        documents : List[str]
            List of text documents to add to the database.
            
        metadatas : List[Dict], optional
            List of metadata dictionaries for each document.
            If None, no metadata will be stored.
            
        ids : List[str], optional
            List of unique identifiers for each document.
            If None, IDs will be auto-generated.
            
        Note:
        -----
        This is an abstract method that must be implemented by subclasses.
        """
        pass

    def query(self, query: str, k: int = 10) -> Dict[str, Any]:
        """
        Search for similar documents in the vector database.
        
        This method should be implemented by concrete subclasses to perform
        semantic search queries against the stored documents.
        
        Parameters:
        -----------
        query : str
            The text query to search for.
            
        k : int, default=10
            Number of most similar documents to return.
            
        Returns:
        --------
        Dict[str, Any]
            Search results containing documents, metadata, and similarity scores.
            The exact structure depends on the implementation.
            
        Note:
        -----
        This is an abstract method that must be implemented by subclasses.
        """
        pass
    
    def delete(self) -> None:
        """
        Delete the vector database collection.
        
        This method should be implemented by concrete subclasses to remove
        the entire collection and all its data.
        
        Note:
        -----
        This is an abstract method that must be implemented by subclasses.
        """
        pass
    

class ChromaVectorDB(_BaseVectorDB):
    """
    ChromaDB implementation of the vector database interface.
    
    This class provides a concrete implementation of the vector database
    using ChromaDB as the backend. ChromaDB is an open-source embedding
    database that supports persistent storage and efficient similarity search.
    
    Features:
    - Persistent storage of document embeddings
    - Efficient similarity search with configurable result count
    - Metadata storage for each document
    - Automatic collection creation if it doesn't exist
    - Support for custom embedding models via LiteLLM
    
    Attributes:
        _client (chromadb.PersistentClient): ChromaDB client instance
        _collection (chromadb.Collection): Active collection for operations
    
    Examples:
    --------
    Basic usage:
    
    ```python
    # Initialize with a new collection
    vector_db = ChromaVectorDB(name="my_documents")
    
    # Add documents
    documents = ["Document 1 content", "Document 2 content"]
    metadatas = [{"source": "file1.txt"}, {"source": "file2.txt"}]
    ids = ["doc1", "doc2"]
    
    vector_db.add(documents, metadatas, ids)
    
    # Search for similar documents
    results = vector_db.query("search query", k=5)
    ```
    
    Using with specific embedding model:
    
    ```python
    # Initialize with OpenAI embeddings
    vector_db = ChromaVectorDB(
        name="research_papers",
        vectorizer_provider="openai",
        vectorizer_model="text-embedding-ada-002"
    )
    ```
    """

    def __init__(self, name: Optional[str] = None, 
                 vectorizer_provider: Optional[str] = None, 
                 vectorizer_model: Optional[str] = None):
        """
        Initialize the ChromaDB vector database.
        
        Parameters:
        -----------
        name : str, optional
            Name of the ChromaDB collection. If provided, the collection
            will be created if it doesn't exist, or connected to if it does.
            
        vectorizer_provider : str, optional
            The embedding provider to use for vectorization.
            Examples: "openai", "anthropic", "cohere"
            
        vectorizer_model : str, optional
            The specific embedding model to use.
            Examples: "text-embedding-ada-002", "text-embedding-3-small"
            
        Examples:
        ---------
        ```python
        # Create new collection
        vector_db = ChromaVectorDB("my_documents")
        
        # Connect to existing collection with custom embeddings
        vector_db = ChromaVectorDB(
            name="existing_collection",
            vectorizer_provider="openai",
            vectorizer_model="text-embedding-ada-002"
        )
        ```
        """
        super().__init__(name, vectorizer_provider, vectorizer_model)
        self._client = chromadb.PersistentClient()
        if name:
            try:
                self._collection = self._client.get_collection(name)
            except chromadb.errors.NotFoundError:
                self._collection = self._client.create_collection(name)

    def add(self, documents: List[str], metadatas: List[Dict], ids: List[str]) -> None:
        """
        Add documents to the ChromaDB collection.
        
        This method adds documents along with their metadata and IDs to the
        ChromaDB collection. The documents are automatically converted to
        embeddings using the configured embedding model.
        
        Parameters:
        -----------
        documents : List[str]
            List of text documents to add to the database.
            Each document will be converted to a vector embedding.
            
        metadatas : List[Dict]
            List of metadata dictionaries for each document.
            Each metadata dict can contain any key-value pairs for
            document categorization and filtering.
            
        ids : List[str]
            List of unique identifiers for each document.
            IDs must be unique within the collection.
            
        Raises:
        -------
        ValueError
            If the lengths of documents, metadatas, and ids don't match.
            
        Examples:
        ---------
        ```python
        # Add documents with metadata
        documents = [
            "Machine learning is a subset of artificial intelligence.",
            "Deep learning uses neural networks with multiple layers."
        ]
        
        metadatas = [
            {"topic": "machine_learning", "source": "textbook", "year": 2023},
            {"topic": "deep_learning", "source": "research_paper", "year": 2023}
        ]
        
        ids = ["doc_001", "doc_002"]
        
        vector_db.add(documents, metadatas, ids)
        ```
        
        Notes:
        ------
        - All three lists must have the same length
        - IDs must be unique within the collection
        - Documents are automatically embedded using the configured model
        - Metadata can be used for filtering during queries
        """
        if not (len(documents) == len(metadatas) == len(ids)):
            raise ValueError("documents, metadatas, and ids must have the same length")
            
        self._collection.add(
            documents=documents,
            metadatas=metadatas,
            ids=ids
        )

    def query(self, query: str, k: int = 10) -> Dict[str, Any]:
        """
        Search for similar documents in the ChromaDB collection.
        
        This method performs semantic search by converting the query to an
        embedding and finding the most similar document embeddings in the
        collection.
        
        Parameters:
        -----------
        query : str
            The text query to search for. This will be converted to a
            vector embedding and compared against stored documents.
            
        k : int, default=10
            Number of most similar documents to return. Higher values
            return more results but may include less relevant documents.
            
        Returns:
        --------
        Dict[str, Any]
            A dictionary containing search results with the following structure:
            {
                'ids': List[List[str]] - Document IDs of retrieved documents,
                'documents': List[List[str]] - The actual document content,
                'metadatas': List[List[Dict]] - Metadata for each document,
                'distances': List[List[float]] - Similarity scores (lower = more similar)
            }
            
        Examples:
        ---------
        ```python
        # Basic search
        results = vector_db.query("What is machine learning?", k=5)
        
        # Access results
        for i, (doc_id, document, metadata, distance) in enumerate(zip(
            results['ids'][0], 
            results['documents'][0], 
            results['metadatas'][0], 
            results['distances'][0]
        )):
            print(f"Result {i+1}:")
            print(f"  ID: {doc_id}")
            print(f"  Content: {document[:100]}...")
            print(f"  Similarity: {1 - distance:.3f}")
            print(f"  Metadata: {metadata}")
        ```
        
        Notes:
        ------
        - Results are returned in order of similarity (most similar first)
        - Distance scores are cosine distances (0 = identical, 2 = opposite)
        - If fewer than k documents exist, all available documents are returned
        - The query is automatically embedded using the same model as stored documents
        """
        results = self._collection.query(
            query_texts=query,
            n_results=k
        )
        return results
    
    
    



from typing import Dict, Union, AsyncGenerator
from ..prompts.prompt import Prompt
from ..prompts.prompt_chain import PromptChain
from ..prompts.iterative_prompt import IterativePrompt
from ..tools._tool_parser import ToolParser
from ..conf import Conf

class PromptExecutorMixin:
    """Mixin class to handle prompt execution."""

    
    async def _execute_stream(self, prompt: Union[str, Prompt, PromptChain]) -> AsyncGenerator[Dict, None]:
        """
        Execute a prompt asynchronously with streaming.
        
        Args:
            prompt: The prompt to execute (string, Prompt, or PromptChain)
            
        Yields:
            Dictionary containing streaming response chunks
        """

        # Handle RAG if available
        if hasattr(self, '_rag') and self._rag:
            response = self._rag.query(prompt)
            if len(response["documents"]) > 0:
                documents = response.get('documents', [])
                documents = [s for doc_list in documents for s in doc_list]
                documents = '\n'.join(documents)
                prompt += Conf()["default_prompt"]["rag"] + documents

        # Handle different prompt types
        if isinstance(prompt, PromptChain):
            async for chunk in self._execute_chain_stream(prompt):
                yield chunk
        elif isinstance(prompt, IterativePrompt):
            async for chunk in self._execute_iterative_stream(prompt):
                yield chunk
        elif isinstance(prompt, Prompt):
            async for chunk in self._completion_stream(str(prompt), response_type=prompt.response_type):
                yield chunk
        else:
            async for chunk in self._completion_stream(prompt):
                yield chunk
    
    def _execute(self, prompt: Union[str, Prompt, PromptChain]) -> Dict:
        """
        Execute a prompt synchronously.
        
        Args:
            prompt: The prompt to execute (string, Prompt, or PromptChain)
            agent: The agent to use for execution
            
        Returns:
            Dictionary containing the response
        """

        if self._rag:
            response = self._rag.query(prompt)
            if len(response["documents"])>0:
                documents = response.get('documents', [])
                documents = [s for doc_list in documents for s in doc_list]
                documents = '\n'.join(documents)
                prompt += Conf()["default_prompt"]["rag"] + documents

        if isinstance(prompt, PromptChain):
            return self._execute_chain(prompt)
        elif isinstance(prompt, IterativePrompt):
            return self._execute_iterative(prompt)
        elif isinstance(prompt, Prompt):
            return self._completion(str(prompt), response_type=prompt.response_type)
        else:
            return self._completion(prompt)

    async def _execute_chain_async(self, chain: PromptChain) -> Dict:
        """
        Execute a prompt chain asynchronously.
        
        Args:
            chain: The prompt chain to execute
            agent: The agent to use for execution
            
        Returns:
            Dictionary containing the final response
        """
        response = None
        for i in range(chain._size):
            current_prompt = chain._format(i, response.output if response else None)
            response = self._completion(current_prompt)
        return response

    async def _execute_chain_stream(self, chain: PromptChain) -> AsyncGenerator[Dict, None]:
        """
        Execute a prompt chain asynchronously with streaming.
        
        Args:
            chain: The promptChain to execute
            
        Yields:
            Streaming response chunks from the final prompt in the chain
        """
        response = None
        for i in range(chain._size):
            current_prompt = chain._format(i, response.output if response else None)
            if i == chain._size - 1:  # Last prompt in chain
                async for chunk in self._completion_stream(current_prompt):
                    yield chunk
            else:  # Execute non-streaming for intermediate prompts
                response = self._completion(current_prompt)

    def _execute_chain(self, chain: PromptChain) -> Dict:
        """
        Execute a prompt chain synchronously.
        
        Args:
            chain: The prompt chain to execute
            agent: The agent to use for execution
            
        Returns:
            Dictionary containing the final response
        """
        response = None
        for i in range(chain._size):
            current_prompt = chain._format(i, response["choices"][0]["message"]["content"] if response else None)
            response = self._completion(current_prompt)
        return response
    
    def _execute_iterative(self, prompt: IterativePrompt) -> Dict:
        """
        Execute an iterative prompt synchronously.
        
        Args:
            prompt: The iterative prompt to execute
            agent: The agent to use for execution
            
        Returns:
            Dictionary containing the final response
        """
        response = ""
        memory = ""
        for i in range(prompt._size):
            if i > 0 and prompt._has_memory:
                if prompt._retain_all:
                    memory += current_response
                else:
                    memory = current_response   
                current_prompt = prompt._format(i, memory)
            else:
                current_prompt = prompt._format(i)
            current_response = self._completion(current_prompt)

            response += current_response
        return response

    async def _execute_iterative_stream(self, prompt: IterativePrompt) -> AsyncGenerator[Dict, None]:
        """
        Execute an iterative prompt asynchronously with streaming.
        
        Args:
            prompt: The iterative prompt to execute
            
        Yields:
            Streaming response chunks from the final iteration
        """
        response = ""
        memory = ""
        for i in range(prompt._size):
            if i > 0 and prompt._has_memory:
                if prompt._retain_all:
                    memory += current_response
                else:
                    memory = current_response   
                current_prompt = prompt._format(i, memory)
            else:
                current_prompt = prompt._format(i)
            
            if i == prompt._size - 1:  # Last iteration
                async for chunk in self._completion_stream(current_prompt):
                    yield chunk
            else:  # Execute non-streaming for intermediate iterations
                current_response = self._completion(current_prompt)
                response += current_response

    def _completion(self, prompt: str|list, response_type: str = None) -> Dict:
        
        from pydantic import BaseModel

        class Response(BaseModel):
            response: response_type

        if response_type!=None:
            response_type = Response

        self._disable_logging()

        url = None
        model = self.provider+"/"+self.model
        
        if hasattr(self, "url") and self.url != None:
            url = self.url+"/v"+str(self.version)
            model = "hosted_vllm/"+model
        
        tools = None

        if hasattr(self, "_tools"):
            tools = []
            tp = ToolParser()
            for tool in self._tools:
                tools.append(tp.parse(tool))      

        if isinstance(prompt, str):
            messages = [{ "content": prompt,"role": "user"}]
        else:
            messages = prompt

        from litellm import completion

        return completion(model=model, 
                          messages=messages, 
                          response_format=response_type,
                          base_url = url,
                          tools=tools,
                          max_tokens=self._max_tokens)


    async def _completion_stream(self, prompt: str|list, response_type: str = None) -> AsyncGenerator[Dict, None]:
        """
        Execute a streaming completion.
        
        Args:
            prompt: The prompt to execute
            response_type: Optional response type
            
        Yields:
            Streaming response chunks
        """
        url = None
        model = self.provider+"/"+self.model
        if hasattr(self, "url") and self.url != None:
            url = self.url+"/v"+str(self.version)
            model = "hosted_vllm/"+self.model
        
        if isinstance(prompt, str):
            messages = [{ "content": prompt,"role": "user"}]
        else:
            messages = prompt

        self._disable_logging()
        from litellm import acompletion

        response = await acompletion(
            model=model, 
            messages=messages, 
            response_format=response_type,
            base_url=url,
            stream=True,
            max_tokens=self._max_tokens
        )
        
        async for chunk in response:
            yield chunk
            

    def _disable_logging(self):
        import logging
        loggers = [
            "LiteLLM Proxy",
            "LiteLLM Router",
            "LiteLLM",
            "httpx"
        ]

        for logger_name in loggers:
            logger = logging.getLogger(logger_name)
            logger.setLevel(logging.CRITICAL + 1) 




import tiktoken
from typing import Union, List, Dict

class TokenCounter:
    _instance = None
    _encoders: Dict[str, tiktoken.Encoding] = {}

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(TokenCounter, cls).__new__(cls)
        return cls._instance

    def _get_encoder(self, model: str) -> tiktoken.Encoding:
        """Get or create an encoder for the specified model."""
        if model not in self._encoders:
            try:
                # Map model names to their corresponding encodings
                if "gpt" in model.lower():
                    self._encoders[model] = tiktoken.encoding_for_model(model)
                elif "claude" in model.lower():
                    self._encoders[model] = tiktoken.get_encoding("cl100k_base")
                elif "llama" in model.lower():
                    self._encoders[model] = tiktoken.get_encoding("cl100k_base")
                else:
                    # Default to cl100k_base for unknown models
                    self._encoders[model] = tiktoken.get_encoding("cl100k_base")
            except Exception as e:
                print(f"Warning: Could not get specific encoder for {model}, using cl100k_base. Error: {e}")
                self._encoders[model] = tiktoken.get_encoding("cl100k_base")
        
        return self._encoders[model]


    def count(self, model: str, input_text: Union[str, List[str]], output_text: Union[str, List[str]]) -> Dict[str, int]:

        encoder = self._get_encoder(model)
        
        if isinstance(input_text, list):
            input_text = "\n".join(input_text)
        if isinstance(output_text, list):
            output_text = "\n".join(output_text)

        input_tokens = len(encoder.encode(input_text))
        output_tokens = len(encoder.encode(output_text))

        return {"input_tokens":input_tokens, "output_tokens":output_tokens, "total_tokens":input_tokens + output_tokens}




if __name__ == "__main__":
    input_text = "Hello, how are you today?"
    output_text = "I'm doing well!"
    model = "gpt-4o-mini"
    print(TokenCounter().count_tokens(model, input_text, output_text))
    


from ._base_model import BaseModel
from ..keys.keys_manager import load_key
from ._response_processor import ResponseProcessorMixin
from ._prompt_executor import PromptExecutorMixin
from typing import Sequence, Dict, Union, AsyncGenerator
from ..tokens.token_counter import TokenCounter
from ..tokens.token_cost import TokenCost
from ..prompts.prompt_chain import PromptChain
from ..prompts.prompt import Prompt
from monoai.conf import Conf

class Model(BaseModel, ResponseProcessorMixin, PromptExecutorMixin):
    """
    Model class for interacting with AI language models.

    This module provides the Model class which serves as the primary interface for interacting
    with various AI language models (like GPT-4, Claude-3, etc.).

    Examples
    --------
    Basic usage:
    ```
    model = Model(provider="openai", model="gpt-4")
    response = model.ask("What is the capital of France?")
    ```

    With prompt:
    ```
    model = Model(
        provider="anthropic",
        model="claude-3",
    )
    prompt = Prompt(
        prompt="What is the capital of {country}?",
        prompt_data={"country": "France"},
        response_type=str
    )
    response = model.ask(prompt)
    ```
    """

    def __init__(
        self, 
        provider: str | None = None, 
        model: str | None = None, 
        system_prompt: str | Sequence[str] = (),
        count_tokens: bool = False, 
        count_cost: bool = False,
        max_tokens: int = None
    ):
        """
        Initialize a new Model instance.

        Parameters
        ----------
        provider : str
            Name of the provider (e.g., 'openai', 'anthropic')
        model : str
            Name of the model (e.g., 'gpt-4', 'claude-3')
        system_prompt : str | Sequence[str], optional
            System prompt or sequence of prompts
        count_tokens : bool, optional
            Whether to count tokens for each request
        count_cost : bool, optional
            Whether to calculate costs for each request
        max_tokens : int, optional
            Maximum number of tokens for each request
        """
        super().__init__(count_tokens, count_cost, max_tokens)
        
        if provider is None:
            provider = Conf()["base_model"]["provider"]
        if model is None:
            model = Conf()["base_model"]["model"]

        load_key(provider)

        self.provider = provider
        self.model = model

    async def _ask_async(self, prompt: Union[str, Prompt, PromptChain]) -> Dict:
        """
        Ask the model asynchronously.

        Parameters
        ----------
        prompt : Union[str, Prompt]
            The prompt to process

        Returns
        -------
        Dict
            Dictionary containing:
            - response: The model's response
            - prompt: The original prompt
            - model: Dictionary with provider and model name
            - tokens: Token counts (if enabled)
            - cost: Cost calculation (if enabled)

        """
        response = await self._execute_async(prompt)
        return self._process_response(
            prompt,
            response,
        )

    
    async def ask_stream(self, prompt: Union[str, Prompt, PromptChain]) -> AsyncGenerator[Dict, None]:
        """
        Ask the model with streaming response.

        Parameters
        ----------
        prompt : Union[str, Prompt, PromptChain]
            The prompt to process

        Yields
        ------
        Dict
            Streaming response chunks
        """
        yield {"provider":self.provider, "model":self.model}
        async for chunk in self._execute_stream(prompt):
            processed_chunk = self._process_chunk(chunk)
            if processed_chunk["delta"] is not None:
                yield processed_chunk



    def ask(self, prompt: Union[str, Prompt, PromptChain]) -> Dict:
        """
        Ask the model.

        Parameters
        ----------
        prompt : Union[str, Prompt]
            The prompt to process

        Returns
        -------
        Dict
            Dictionary containing:
            - response: The model's response
            - prompt: The original prompt
            - model: Dictionary with provider and model name
            - tokens: Token counts (if enabled)
            - cost: Cost calculation (if enabled)

        """

        response = self._execute(prompt)
        return self._process_response(
            prompt,
            response
        )


"""
This module provides classes for handling different types of prompts in the COICOI framework.
Prompt can be provided as a string or as a xml-like .prompt file.
"""

from .prompt_chain import PromptChain
from .prompt import Prompt  
from .iterative_prompt import IterativePrompt

__all__ = ['Prompt', 'PromptChain', 'IterativePrompt']



import os
import json
import uuid
import sqlite3
from monoai.models import Model
import datetime
from pymongo import MongoClient

class BaseHistory:

    def __init__(self, 
                 path: str, 
                 last_n: int=None): 
        self._history_path = path
        self._last_n = last_n
        
    def generate_chat_id(self):
        return str(uuid.uuid4())

    def load(self):
        pass

    def store(self, chat_id: str, messages: list):
        # Aggiungi un timestamp ISO 8601 a ciascun messaggio
        now = datetime.datetime.utcnow().isoformat() + 'Z'
        for msg in messages:
            if 'timestamp' not in msg:
                msg['timestamp'] = now
        return messages

    def clear(self):
        pass


class JSONHistory(BaseHistory):
    
    def __init__(self, 
                 path, 
                 last_n: int=None): 
        self._history_path = path
        self._last_n = last_n
        if not os.path.exists(self._history_path):
            os.makedirs(self._history_path)

    def load(self, chat_id: str):
        # Ensure proper path construction with os.path.join
        file_path = os.path.join(self._history_path, chat_id + ".json")
        with open(file_path, "r") as f:
            self.messages = json.load(f)
        if self._last_n is not None and len(self.messages) > (self._last_n+1)*2:
            self.messages = [self.messages[0]]+self.messages[-self._last_n*2:]
        return self.messages
    
    def new(self, system_prompt: str):
        chat_id = self.generate_chat_id()
        # Ensure directory exists before storing
        if not os.path.exists(self._history_path):
            os.makedirs(self._history_path, exist_ok=True)
        self.store(chat_id, [{"role": "system", "content": system_prompt}])
        return chat_id

    def store(self, chat_id: str, messages: list):
        messages = super().store(chat_id, messages)
        # Load existing messages
        file_path = os.path.join(self._history_path, chat_id + ".json")
        try:
            with open(file_path, "r") as f:
                existing_messages = json.load(f)
        except FileNotFoundError:
            existing_messages = []
        
        # Add the new messages (già con timestamp)
        new_messages = existing_messages + messages
        
        with open(file_path, "w") as f:
            json.dump(new_messages, f, indent=4)

class SQLiteHistory(BaseHistory):
    
    def __init__(self, path: str="histories/chat.db", last_n: int=None):
        self._db_path = path
        self._last_n = last_n
        self._init_db()
    
    def _init_db(self):
        os.makedirs(os.path.dirname(self._db_path), exist_ok=True)
        with sqlite3.connect(self._db_path) as conn:
            conn.execute("""
                CREATE TABLE IF NOT EXISTS messages (
                    chat_id TEXT,
                    order_index INTEGER,
                    role TEXT,
                    content TEXT,
                    PRIMARY KEY (chat_id, order_index)
                )
            """)
    
    def load(self, chat_id: str):
        with sqlite3.connect(self._db_path) as conn:
            if self._last_n is not None:
                # Get system message
                cursor = conn.execute(
                    "SELECT role, content FROM messages WHERE chat_id = ? AND order_index = 0",
                    (chat_id,)
                )
                system_message = cursor.fetchone()
                
                # Get last N messages
                cursor = conn.execute(
                    """
                    SELECT role, content 
                    FROM messages 
                    WHERE chat_id = ? 
                    ORDER BY order_index DESC 
                    LIMIT ?
                    """,
                    (chat_id, self._last_n * 2)
                )
                last_messages = [{"role": role, "content": content} for role, content in cursor]
                last_messages.reverse()  # Reverse to get correct order
                
                # Combine system message with last N messages
                self.messages = [{"role": system_message[0], "content": system_message[1]}] + last_messages
            else:
                cursor = conn.execute(
                    "SELECT role, content FROM messages WHERE chat_id = ? ORDER BY order_index",
                    (chat_id,)
                )
                self.messages = [{"role": role, "content": content} for role, content in cursor]
        return self.messages
    
    def new(self, system_prompt: str):
        chat_id = self.generate_chat_id()
        self.store(chat_id, [{"role": "system", "content": system_prompt, "order_index": 0}])
        return chat_id

    def store(self, chat_id: str, messages: list):
        messages = super().store(chat_id, messages)
        with sqlite3.connect(self._db_path) as conn:
            # Get the last order_index
            cursor = conn.execute(
                "SELECT MAX(order_index) FROM messages WHERE chat_id = ?",
                (chat_id,)
            )
            last_index = cursor.fetchone()[0]
            
            # If no messages exist yet, start from -1
            if last_index is None:
                last_index = -1
            
            # Insert the new messages con timestamp
            for i, message in enumerate(messages, start=last_index + 1):
                conn.execute(
                    "INSERT INTO messages (chat_id, order_index, role, content) VALUES (?, ?, ?, ?)",
                    (chat_id, i, message["role"], message["content"])
                )
                conn.commit()
        

class DictHistory(BaseHistory):
    """
    In-memory history storage using Python dictionaries.
    Useful for testing and temporary conversations.
    """
    
    def __init__(self, last_n: int = None):
        self._last_n = last_n
        self._histories = {}  # Dictionary to store chat histories
    
    def load(self, chat_id: str):
        if chat_id not in self._histories:
            self.messages = []
            return self.messages
        
        messages = self._histories[chat_id]
        if self._last_n is not None and len(messages) > (self._last_n + 1) * 2:
            messages = [messages[0]] + messages[-self._last_n * 2:]
        
        self.messages = messages
        return self.messages
    
    def new(self, system_prompt: str):
        chat_id = self.generate_chat_id()
        messages = [{"role": "system", "content": system_prompt}]
        self.store(chat_id, messages)
        return chat_id
    
    def store(self, chat_id: str, messages: list):
        messages = super().store(chat_id, messages)
        
        if chat_id not in self._histories:
            self._histories[chat_id] = []
        
        # Add the new messages (già con timestamp)
        self._histories[chat_id].extend(messages)
    
    def clear(self, chat_id: str):
        """Clear history for a specific chat."""
        if chat_id in self._histories:
            del self._histories[chat_id]
    
    def clear_all(self):
        """Clear all chat histories."""
        self._histories.clear()
    
    def get_all_chat_ids(self):
        """Get all chat IDs currently stored."""
        return list(self._histories.keys())
    
    def get_chat_count(self):
        """Get the total number of chats stored."""
        return len(self._histories)


class MongoDBHistory(BaseHistory):
    def __init__(self, db_path, db_name: str = "chat", collection_name: str = "histories", last_n: int = None):
        self._uri = db_path
        self._db_name = db_name
        self._collection_name = collection_name
        self._last_n = last_n
        self._client = MongoClient(self._uri)
        self._db = self._client[self._db_name]
        self._collection = self._db[self._collection_name]

    def load(self, chat_id: str):
        doc = self._collection.find_one({"chat_id": chat_id})
        if not doc:
            self.messages = []
            return self.messages
        messages = doc.get("messages", [])
        if self._last_n is not None and len(messages) > (self._last_n + 1) * 2:
            messages = [messages[0]] + messages[-self._last_n * 2:]
        self.messages = messages
        return self.messages

    def new(self, system_prompt: str):
        chat_id = self.generate_chat_id()
        messages = [{"role": "system", "content": system_prompt}]
        self.store(chat_id, messages)
        return chat_id

    def store(self, chat_id: str, messages: list):
        messages = super().store(chat_id, messages)
        # Get existing messages
        doc = self._collection.find_one({"chat_id": chat_id})
        existing_messages = doc.get("messages", []) if doc else []
        
        # Add the new messages (già con timestamp)
        new_messages = existing_messages + messages
        
        self._collection.update_one(
            {"chat_id": chat_id},
            {"$set": {"messages": new_messages}},
            upsert=True
        )


class HistorySummarizer():

    def __init__(self, model: Model, max_tokens: int=None):
        self._model = model
        self._max_tokens = max_tokens

    def summarize(self, messages: list):
        response = self._model.ask("Summarize the following conversation: "+json.dumps(messages))
        response = response["response"]
        return response




"""
Models are the core of MonoAI. They are responsible for executing prompts and returning responses.

This package uses lazy loading to avoid importing heavy optional dependencies
at module import time. Classes are imported only when accessed.
"""

__all__ = ['Model', "HostedModel", 'MultiModel', 'CollaborativeModel', 'ImageModel', 'VoiceModel']

def __getattr__(name):
	if name == 'Model':
		from .model import Model
		return Model
	elif name == 'HostedModel':
		from .hosted_model import HostedModel
		return HostedModel
	elif name == 'MultiModel':
		from .multi_model import MultiModel
		return MultiModel
	elif name == 'CollaborativeModel':
		from .collaborative_model import CollaborativeModel
		return CollaborativeModel
	elif name == 'ImageModel':
		from .image_model import ImageModel
		return ImageModel
	elif name == 'VoiceModel':
		from .voice_model import VoiceModel
		return VoiceModel
	raise AttributeError(f"module 'monoai.models' has no attribute {name!r}")


import os
from pathlib import Path
from typing import Any, Dict, Optional, Union
import yaml
from yaml.parser import ParserError

class Conf:
    """
    A singleton class for managing configuration in a programmatic way.
    Configuration defined here overrides the configuration in monoai.yaml file.

    Examples
    --------
    ```
    from monoai.conf import Conf
    
    # override the base model
    Conf()["base_model"] = {
        "provider": "openai",
        "model": "gpt-4o-mini"
    }
    
    # set prompts path programmatically
    Conf()["prompts_path"] = "prompts"
    ```
    """
    
    _instance = None
    _DEFAULT_CONFIG = {
        'keysfile_path': "providers.keys",
        'supported_files': {
            "text": ["txt", "py", "md"],
            "image": ["png", "jpg", "jpeg", "gif", "webp"]
        },
        'prompts_path': "",
        'default_prompt': {
            "rag": "Use also the following information to answer the question: ",
            "summary":"Here is the summary of the conversation so far:\n\n",
            "file":"Here is the content of the file:\n\n"
        }
    }
    
    def __new__(cls):
        """
        Create or return the singleton instance.

        Returns
        -------
        Config
            The singleton instance
        """
        if cls._instance is None:
            cls._instance = super(Conf, cls).__new__(cls)
            cls._instance._initialize()
        return cls._instance
    
    def _initialize(self):
        """
        Initialize the Config instance.
        
        Loads and validates the configuration file, setting up defaults
        and performing environment variable interpolation.
        """
        self._config_path = Path('monoai.yaml')
        self._config = self._DEFAULT_CONFIG.copy()
        self._load_config()
    
    def _load_config(self) -> None:
        """
        Load configuration from the YAML file.
        
        Handles file reading, YAML parsing, environment variable interpolation,
        and configuration validation.

        Raises
        ------
        FileNotFoundError
            If the configuration file doesn't exist
        yaml.ParserError
            If the YAML syntax is invalid
        ValueError
            If the configuration structure is invalid
        """
        try:
            if self._config_path.exists():
                with open(self._config_path, 'r') as f:
                    file_config = yaml.safe_load(f)
                if file_config:
                    self._merge_config(self._config, file_config)            
            
        except ParserError as e:
            raise ValueError(f"Invalid YAML syntax in {self._config_path}: {e}")
        except Exception as e:
            raise ValueError(f"Error loading configuration: {e}")
    
    def _merge_config(self, base: Dict, override: Dict) -> None:
        """
        Recursively merge override configuration into base configuration.

        Parameters
        ----------
        base : Dict
            The base configuration dictionary
        override : Dict
            The override configuration dictionary
        """
        for key, value in override.items():
            if (
                key in base and 
                isinstance(base[key], dict) and 
                isinstance(value, dict)
            ):
                self._merge_config(base[key], value)
            else:
                base[key] = value
        
        
    def __getitem__(self, key: str) -> Any:
        return self._config[key]
    
    def __setitem__(self, key: str, value: Any) -> None:
        self._config[key] = value
    
    
    
    


from ._base_model import BaseModel
from .model import Model
from .multi_model import MultiModel
from ._prompt_executor import PromptExecutorMixin
from ._response_processor import ResponseProcessorMixin
from typing import List, Dict, Union
import asyncio
from ..prompts.prompt_chain import PromptChain
from ..prompts.prompt import Prompt

class CollaborativeModel(BaseModel, PromptExecutorMixin, ResponseProcessorMixin):
    """
    A class to implement collaborative decision making across multiple AI models.
    
    CollaborativeModel manages a collection of AI models and an aggregator model.
    It executes prompts across all models in parallel and then uses the aggregator
    to synthesize a final response based on all individual responses.

    Examples
    --------
    Basic collaborative analysis:
    ```
    models = [
        {"provider": "openai", "model": "gpt-4"},
        {"provider": "anthropic", "model": "claude-3"}
    ]
    aggregator = {"provider": "openai", "model": "gpt-4"}
    collab = CollaborativeModel(models=models, aggregator=aggregator)
    response = collab.ask("Explain quantum computing")
    print(response["response"])  # Aggregated response
    for ind_resp in response["individual_responses"]:
        print(f"{ind_resp['model']['name']}: {ind_resp['response']}")
    ```
    """

    def __init__(
        self,
        models: List[Dict[str, str]],
        aggregator: Dict[str, str],
        count_tokens: bool = False,
        count_cost: bool = False
    ):
        """
        Initialize a new CollaborativeModel instance.

        Parameters
        ----------
        models : List[Dict[str, str]]
            List of dictionaries with provider and model information
        aggregator : Dict[str, str]
            Dictionary with provider and model information for the aggregator
        count_tokens : bool, optional
            Whether to count tokens for each request
        count_cost : bool, optional
            Whether to calculate costs for each request
        """
        super().__init__(count_tokens, count_cost)

        self._multi_model = MultiModel(
            models=models,
            count_tokens=count_tokens,
            count_cost=count_cost
        )

        self._aggregator = Model(
            aggregator['provider'],
            aggregator['model'],
            count_tokens=count_tokens,
            count_cost=count_cost
        )

    def _format_aggregator_prompt(self, prompt: Union[str, Prompt, PromptChain], responses: List[Dict]) -> str:
        """
        Format the prompt for the aggregator model.

        Parameters
        ----------
        prompt : Union[str, Prompt, PromptChain]
            The original prompt
        responses : List[Dict]
            List of responses from individual models

        Returns
        -------
        str
            Formatted prompt for the aggregator including original question
            and all model responses
        """
        prompt_text = str(prompt)
        model_responses = "\n\n".join([
            f"Model {i+1} ({response['model']['provider']} - {response['model']['name']}):\n{response['response']}"
            for i, response in enumerate(responses)
        ])
        
        return f"""Please analyze the following responses from different models and provide a comprehensive answer:
                    Original Question: {prompt_text}
                    Model Responses:
                    {model_responses}
                    Please provide a well-reasoned response that takes into account all the information above."""

    async def _ask_async(self, prompt: Union[str, Prompt, PromptChain]) -> Dict:
        """
        Ask all models and aggregate their responses asynchronously.

        Parameters
        ----------
        prompt : Union[str, Prompt, PromptChain]
            The prompt to process across all models

        Returns
        -------
        Dict
            Dictionary containing:
            - response: The aggregated response
            - prompt: The original prompt
            - model: Dictionary with aggregator's provider and model name
            - tokens: Token counts (if enabled)
            - cost: Cost calculation (if enabled)
            - individual_responses: List of responses from individual models

        Examples
        --------
        Using async/await:
            >>> response = await collab.ask_async("What is consciousness?")
            >>> print(response["response"])  # Aggregated response
            >>> for resp in response["individual_responses"]:
            ...     print(f"{resp['model']['name']}: {resp['response']}")
        """
        # Get responses from all models
        model_responses = await self._multi_model.ask_async(prompt)
        
        # Get aggregator response
        aggregator_prompt = self._format_aggregator_prompt(prompt, model_responses)
        aggregator_response = await self._execute_async(aggregator_prompt, self._aggregator._agent)
        
        # Process aggregator response
        processed_aggregator = self._process_response(
            aggregator_prompt,
            aggregator_response,
            self._aggregator.provider,
            self._aggregator.model,
            self._count_tokens,
            self._count_cost
        )

        processed_aggregator["individual_responses"] = model_responses
        return processed_aggregator

    def ask(self, prompt: Union[str, Prompt, PromptChain]) -> Dict:
        """
        Ask all models and aggregate their responses synchronously.

        Parameters
        ----------
        prompt : Union[str, Prompt, PromptChain]
            The prompt to process across all models

        Returns
        -------
        Dict
            Dictionary containing:
            - response: The aggregated response
            - prompt: The original prompt
            - model: Dictionary with aggregator's provider and model name
            - tokens: Token counts (if enabled)
            - cost: Cost calculation (if enabled)
            - individual_responses: List of responses from individual models

        """
        return asyncio.run(self.ask_async(prompt))




from ..models import Model
from .agentic_loop import (
    FunctionCallingAgenticLoop, 
    ReactAgenticLoop, 
    ReactWithFCAgenticLoop,
    PlanAndExecuteAgenticLoop,
    ProgrammaticAgenticLoop,
    ReflexionAgenticLoop,
    SelfAskAgenticLoop,
    SelfAskWithSearchLoop,
    _AgenticLoop
)

from ..prompts import Prompt


class Agent:
    """Agente AI che implementa diversi paradigmi di ragionamento.
    
    Questa classe fornisce un'interfaccia unificata per creare e utilizzare
    agenti AI con diversi paradigmi di ragionamento. L'agente può essere
    configurato per utilizzare function calling, React, plan-and-execute,
    e altri approcci agentici, oppure un paradigma personalizzato.
    
    Attributes
    ----------
    _model : Model
        Modello AI da utilizzare per l'esecuzione
    _loop : Any
        Istanza del loop agentico specifico per il paradigma scelto
    """
    
    def __init__(self, model: Model, tools=None, paradigm="function_calling", 
                 agent_prompt=None, debug=False, max_iter=None):
        """Inizializza l'agente con il modello e la configurazione specificata.
        
        Parameters
        ----------
        model : Model
            Modello AI da utilizzare per l'esecuzione dell'agente
        tools : list, optional
            Lista degli strumenti disponibili per l'agente. Default è None.
        paradigm : str or _AgenticLoop, optional
            Paradigma di ragionamento da utilizzare. Può essere:
            
            **Stringhe predefinite:**
            - "function_calling": Utilizza OpenAI function calling
            - "react": Approccio React standard
            - "react_with_function_calling": Combina React e function calling
            - "plan-and-execute": Paradigma plan-and-execute
            - "programmatic": Approccio programmatico
            - "reflexion": Paradigma con riflessione
            - "self_ask": Paradigma self-ask
            - "self_ask_with_search": Paradigma self-ask con ricerca web
            
            **Oggetto personalizzato:**
            - Un'istanza di una classe derivata da _AgenticLoop
            
            Default è "function_calling".
        agent_prompt : str, optional
            Prompt personalizzato per l'agente. Se None, viene utilizzato
            il prompt di default per il paradigma scelto. Default è None.
        debug : bool, optional
            Flag per abilitare la stampa di debug durante l'esecuzione.
            Default è False.
        max_iter : int, optional
            Numero massimo di iterazioni consentite per l'agente.
            Se None, non ci sono limiti. Default è None.
        
        Raises
        ------
        ValueError
            Se il paradigma specificato non è supportato o non è valido
        TypeError
            Se viene passato un oggetto personalizzato che non deriva da _AgenticLoop
        """
        self._model = model

        # Gestione paradigma personalizzato
        if isinstance(paradigm, _AgenticLoop):
            # Verifica che l'oggetto personalizzato sia valido
            if not hasattr(paradigm, 'start') or not callable(paradigm.start):
                raise TypeError("Il paradigma personalizzato deve avere un metodo 'start' callable")
            self._loop = paradigm
        else:
            # Paradigmi predefiniti
            loop_kwargs = self._model, agent_prompt, debug, max_iter
            
            if paradigm == "function_calling":
                self._loop = FunctionCallingAgenticLoop(*loop_kwargs)
            elif paradigm == "react":
                self._loop = ReactAgenticLoop(*loop_kwargs)
            elif paradigm == "react_with_function_calling":
                self._loop = ReactWithFCAgenticLoop(*loop_kwargs)
            elif paradigm == "plan-and-execute":
                self._loop = PlanAndExecuteAgenticLoop(*loop_kwargs)
            elif paradigm == "programmatic":
                self._loop = ProgrammaticAgenticLoop(*loop_kwargs)
            elif paradigm == "reflexion":
                self._loop = ReflexionAgenticLoop(*loop_kwargs)
            elif paradigm == "self_ask":
                self._loop = SelfAskAgenticLoop(*loop_kwargs)
            elif paradigm == "self_ask_with_search":
                self._loop = SelfAskWithSearchLoop(*loop_kwargs)
            else:
                raise ValueError(f"Paradigma '{paradigm}' non supportato. "
                               f"Paradigmi disponibili: function_calling, react, "
                               f"react_with_function_calling, plan-and-execute, "
                               f"programmatic, reflexion, self_ask, self_ask_with_search, "
                               f"oppure un oggetto personalizzato derivato da _AgenticLoop")
        
        if tools is not None:
            self._loop.register_tools(tools)
        
    def run(self, prompt: str | Prompt):
        """Esegue l'agente con il prompt specificato.
        
        Parameters
        ----------
        prompt : str
            Prompt o query dell'utente da processare
        
        Returns
        -------
        Dict[str, Any]
            Risposta dell'agente che include:
            - prompt: Query originale
            - iterations: Lista delle iterazioni processate
            - response: Risposta finale (se presente)
        
        Notes
        -----
        Questo metodo delega l'esecuzione al loop agentico specifico
        configurato durante l'inizializzazione. Il comportamento esatto
        dipende dal paradigma scelto (predefinito o personalizzato).
        """
        
        return self._loop.start(prompt)


from .model import Model
from ._response_processor import ResponseProcessorMixin
from ._prompt_executor import PromptExecutorMixin
from typing import Sequence, Dict, Union
from ..keys.keys_manager import load_key


class HostedModel(Model, ResponseProcessorMixin, PromptExecutorMixin):
    """
    Model class for interacting with AI language models.

    This module provides the Model class which serves as the primary interface for interacting
    with various AI language models (like GPT-4, Claude-3, etc.).

    Examples
    --------
    Basic usage:
    ```
    model = Model(provider="openai", model="gpt-4")
    response = model.ask("What is the capital of France?")
    ```

    With prompt:
    ```
    model = Model(
        provider="anthropic",
        model="claude-3",
    )
    prompt = Prompt(
        prompt="What is the capital of {country}?",
        prompt_data={"country": "France"},
        response_type=str
    )
    response = model.ask(prompt)
    ```
    """

    def __init__(
        self, 
        url: str,
        version: int = 1,
        provider: str | None = None, 
        model: str | None = None, 
        system_prompt: str | Sequence[str] = (),
        count_tokens: bool = False, 
        count_cost: bool = False,
        max_tokens: int = None
    ):
        """
        Initialize a new Model instance.

        Parameters
        ----------
        provider : str
            Name of the provider (e.g., 'openai', 'anthropic')
        model : str
            Name of the model (e.g., 'gpt-4', 'claude-3')
        system_prompt : str | Sequence[str], optional
            System prompt or sequence of prompts
        count_tokens : bool, optional
            Whether to count tokens for each request
        count_cost : bool, optional
            Whether to calculate costs for each request
        max_tokens : int, optional
            Maximum number of tokens for each request
        """

        super().__init__(
            provider,
            model,
            system_prompt, 
            count_tokens, 
            count_cost, 
            max_tokens
        )
        
        self.url = url
        self.version = version




import pytest
import tempfile
import os
from monoai.rag.documents_builder import DocumentsBuilder


def test_custom_split_function():
    """Test that custom split functions work correctly."""
    
    def custom_split_by_lines(text, chunk_size, chunk_overlap):
        """Custom split function that splits by lines."""
        lines = text.split('\n')
        chunks = []
        for i in range(0, len(lines), chunk_size - chunk_overlap):
            chunk_lines = lines[i:i + chunk_size]
            if chunk_lines:  # Only add non-empty chunks
                chunks.append('\n'.join(chunk_lines))
        return chunks
    
    # Create test text with multiple lines
    test_text = "Line 1\nLine 2\nLine 3\nLine 4\nLine 5\nLine 6\nLine 7\nLine 8"
    
    # Create DocumentsBuilder with custom split function
    builder = DocumentsBuilder(
        chunk_strategy="custom",
        chunk_size=3,  # 3 lines per chunk
        chunk_overlap=1,  # 1 line overlap
        custom_split_func=custom_split_by_lines
    )
    
    # Test the custom split
    chunks = builder._split_text(test_text)
    
    # Expected chunks: ["Line 1\nLine 2\nLine 3", "Line 3\nLine 4\nLine 5", "Line 5\nLine 6\nLine 7", "Line 7\nLine 8"]
    expected_chunks = [
        "Line 1\nLine 2\nLine 3",
        "Line 3\nLine 4\nLine 5", 
        "Line 5\nLine 6\nLine 7",
        "Line 7\nLine 8"
    ]
    
    assert chunks == expected_chunks


def test_custom_split_with_from_str():
    """Test that custom split works with from_str method."""
    
    def custom_split_by_sections(text, chunk_size, chunk_overlap):
        """Custom split function that splits by section markers."""
        sections = text.split('---')
        chunks = []
        for i in range(0, len(sections), chunk_size - chunk_overlap):
            chunk_sections = sections[i:i + chunk_size]
            if chunk_sections:  # Only add non-empty chunks
                chunks.append('---'.join(chunk_sections))
        return chunks
    
    test_text = "Section 1---Section 2---Section 3---Section 4---Section 5"
    
    builder = DocumentsBuilder(
        chunk_strategy="custom",
        chunk_size=2,  # 2 sections per chunk
        chunk_overlap=0,  # No overlap
        custom_split_func=custom_split_by_sections
    )
    
    chunks, metadata, ids = builder.from_str(test_text)
    
    expected_chunks = [
        "Section 1---Section 2",
        "Section 3---Section 4", 
        "Section 5"
    ]
    
    assert chunks == expected_chunks
    assert len(metadata) == len(chunks)
    assert len(ids) == len(chunks)


def test_custom_split_validation():
    """Test validation of custom split function parameters."""
    
    # Test that custom_split_func is required when strategy is "custom"
    with pytest.raises(ValueError, match="custom_split_func must be provided"):
        DocumentsBuilder(chunk_strategy="custom")
    
    # Test that custom_split_func must be callable
    with pytest.raises(ValueError, match="custom_split_func must be callable"):
        DocumentsBuilder(
            chunk_strategy="custom",
            custom_split_func="not a function"
        )
    
    # Test that custom_split_func can be None for other strategies
    builder = DocumentsBuilder(chunk_strategy="word", custom_split_func=None)
    assert builder._custom_split_func is None


def test_custom_split_function_signature():
    """Test that custom split function receives correct parameters."""
    
    def test_split_func(text, chunk_size, chunk_overlap):
        """Test function that records received parameters."""
        test_split_func.called = True
        test_split_func.text = text
        test_split_func.chunk_size = chunk_size
        test_split_func.chunk_overlap = chunk_overlap
        return [text]  # Return single chunk
    
    test_split_func.called = False
    
    builder = DocumentsBuilder(
        chunk_strategy="custom",
        chunk_size=100,
        chunk_overlap=20,
        custom_split_func=test_split_func
    )
    
    test_text = "This is a test text"
    builder._split_text(test_text)
    
    assert test_split_func.called
    assert test_split_func.text == test_text
    assert test_split_func.chunk_size == 100
    assert test_split_func.chunk_overlap == 20


def test_auto_custom_strategy():
    """Test that chunk_strategy is automatically set to 'custom' when custom_split_func is provided."""
    
    def dummy_split_func(text, chunk_size, chunk_overlap):
        return [text]
    
    # Test with explicit "word" strategy but custom function
    builder = DocumentsBuilder(
        chunk_strategy="word",
        chunk_size=100,
        chunk_overlap=10,
        custom_split_func=dummy_split_func
    )
    
    # Strategy should be automatically set to "custom"
    assert builder._chunk_strategy == "custom"
    
    # Test with no strategy specified but custom function
    builder2 = DocumentsBuilder(
        chunk_size=100,
        chunk_overlap=10,
        custom_split_func=dummy_split_func
    )
    
    # Strategy should be automatically set to "custom"
    assert builder2._chunk_strategy == "custom"
    
    # Test that normal strategies work when no custom function is provided
    builder3 = DocumentsBuilder(chunk_strategy="word")
    assert builder3._chunk_strategy == "word"


if __name__ == "__main__":
    pytest.main([__file__]) 


import requests
import nbformat
from nbconvert import PythonExporter

def colab_downloader(colab_url):
    """
    Downloads a Jupyter notebook from Google Drive and returns its Python code as a string.

    Args:
        colab_url (str): The Google Drive URL of the notebook (should contain 'drive/').
    """
    
    BASE_URL = "https://docs.google.com/uc?export=download&id="
    try:
        file_id = colab_url.split("drive/")[1].split("?")[0]
    except Exception:
        return "Error: Invalid Google Drive URL."

    download_url = BASE_URL + file_id

    try:
        response = requests.get(download_url)
        response.raise_for_status()
        notebook_content = response.content
    except Exception as e:
        return f"Error downloading notebook: {e}"

    try:
        nb = nbformat.reads(notebook_content.decode('utf-8'), as_version=4)
        exporter = PythonExporter()
        source, _ = exporter.from_notebook_node(nb)
        return source
    except Exception as e:
        return f"Error converting notebook: {e}"






from openai import OpenAI
from ..keys.keys_manager import load_key

class ImageModel:
    """
    A class to interact with AI image generation models.
    
    ImageModel provides an interface for generating images from text prompts using
    AI models. Currently supports OpenAI's DALL-E 3, with potential for expansion
    to other providers and models in the future.

    Examples
    --------
    Basic image generation:
    ```
    model = ImageModel(provider="openai", model="dall-e-3")
    response = model.generate("A beautiful garden with flowers")
    ```
    """

    def __init__(self, provider: str, model: str):
        """
        Initialize a new ImageModel instance.

        Parameters
        ----------
        provider : str
            Name of the provider (currently only "openai" is supported)
        model : str
            Name of the model (currently only "dall-e-3" is supported)

        Raises
        ------
        ValueError
            If an unsupported provider or model is specified
        """
        self.provider = provider
        self.model = model

        if provider.lower() != "openai":
            raise ValueError(f"Provider {provider} not supported")
        if model.lower() != "dall-e-3":
            raise ValueError(f"Model {model} not supported")
        
        load_key(provider)
        self._client = OpenAI()

    def generate(self, 
                prompt: str, 
                size: str = "1024x1024", 
                quality: str = "standard", 
                n: int = 1) -> dict:
        """
        Generate images from a text prompt.

        Parameters
        ----------
        prompt : str
            The text description of the image to generate
        size : str, optional
            The size of the generated image(s). Options:
            - "1024x1024" (default)
            - "1792x1024"
            - "1024x1792"
        quality : str, optional
            The quality of the generated image(s). Options:
            - "standard" (default)
            - "hd"
        n : int, optional
            Number of images to generate (default: 1)

        Returns
        -------
        dict
            OpenAI image generation response containing:
            - created: timestamp
            - data: list of generated images with URLs and other metadata
        """
        response = self._client.images.generate(
            model=self.model,
            prompt=prompt,
            size=size,
            quality=quality,
            n=n,
        )
        return response






import pytest
import shutil
import os
import time
from pathlib import Path
from monoai.chat import Chat
from monoai.models import Model

TEST_MODEL = "gpt-4o-nano"
TEST_SYSTEM_PROMPT = "You are a calculator, return only the number without any any other text"

"""
@pytest.fixture(autouse=True)
def cleanup_histories():
    yield
    # Delete JSON history folder
    json_history_path = Path("histories")
    if json_history_path.exists():
        shutil.rmtree(json_history_path)
    
    # Delete SQLite history folder
    sqlite_history_path = Path("sqlite_histories")
    if sqlite_history_path.exists():
        shutil.rmtree(sqlite_history_path)
"""
        
@pytest.fixture
def json_chat():
    """Create a new chat with JSON history."""
    model = Model(provider="openai", model=TEST_MODEL)
    return Chat(model=model, system_prompt=TEST_SYSTEM_PROMPT)

@pytest.fixture
def sqlite_chat():
    """Create a new chat with SQLite history."""
    model = Model(provider="openai", model=TEST_MODEL)
    return Chat(model=model, system_prompt=TEST_SYSTEM_PROMPT, history_type="sqlite")

def test_chat_initialization():
    """Test that chat is properly initialized with required attributes."""
    model = Model(provider="openai", model=TEST_MODEL)
    chat = Chat(model=model, system_prompt=TEST_SYSTEM_PROMPT)
    assert chat.chat_id is not None
    assert chat._history is not None

def test_json_history_basic_operations(json_chat):
    """Test basic operations with JSON history."""
    # Test initial calculation
    response = json_chat.ask("2+2")
    assert response == "4"
    
    # Test continuation
    response = json_chat.ask("+2")
    assert response == "6"

def test_json_history_persistence(json_chat):
    """Test that JSON history persists between chat instances."""
    # First chat instance
    json_chat.ask("2+2")
    json_chat.ask("+2")
    
    # New chat instance with same history
    model = Model(provider="openai", model=TEST_MODEL)
    new_chat = Chat(model=model, history_type="json", chat_id=json_chat.chat_id)
    response = new_chat.ask("+2")
    assert response == "8"

def test_sqlite_history_basic_operations(sqlite_chat):
    """Test basic operations with SQLite history."""
    # Test initial calculation
    response = sqlite_chat.ask("2+2")
    assert response == "4"
    
    # Test continuation
    response = sqlite_chat.ask("+2")
    assert response == "6"

def test_sqlite_history_persistence(sqlite_chat):
    """Test that SQLite history persists between chat instances."""
    # First chat instance
    sqlite_chat.ask("2+2")
    sqlite_chat.ask("+2")


def test_summarizer():
    """Test that the summarizer works."""
    model = Model(provider="openai", model=TEST_MODEL)
    summarizer_model = Model(provider="openai", model="gpt-4o-mini")
    new_chat = Chat(model=model, 
                    history_type="json", 
                    history_summarizer_model=summarizer_model, 
                    history_summarizer_max_tokens=100)
    
    new_chat.ask("Mi chiamo Giuseppe")
    response = new_chat.ask("Come mi chiamo? Ritorna solo il mio nome e niente altro.")
    assert response.lower() == "giuseppe"



import json
from typing import Dict
from ..tokens.token_counter import TokenCounter
from ..tokens.token_cost import TokenCost
from ..prompts.prompt import Prompt

class ResponseProcessorMixin:
    def _process_response(
        self,
        prompt: Prompt,
        response: Dict
    ) -> Dict:
        """
        Process the response and add optional token and cost information.
        
        Args:
            question: The input question
            answer: The model's answer
            provider: Name of the provider
            model: Name of the model
            count_tokens: Whether to count tokens
            count_cost: Whether to calculate costs
            
        Returns:
            Dictionary containing the response and optional stats
        """
        
        if isinstance(prompt, Prompt):
            if prompt.response_type != None:
                response = json.loads(response)["response"]

        processed_response = {
            "prompt": str(prompt), 
            "response": response["choices"][0]["message"]["content"],
            #"messages_trace": response.all_messages(),
            "model": {
                "provider": self.provider, 
                "name": self.model
            }
        }

        if self._count_tokens:
            processed_response["usage"]={
                "completion_tokens":response["usage"]["completion_tokens"],
                "prompt_tokens":response["usage"]["prompt_tokens"],
                "total_tokens=": response["usage"]["total_tokens"]
                }
                        
        return processed_response 


    def _process_chunk(self, chunk):
        delta = chunk["choices"][0]["delta"]["content"]
        return {"delta":delta}


from ..models import MultiModel
from typing import List, Dict

class MultiAgent(MultiModel):

    def __init__(self, models: List[Dict[str, str]], tools):
        super().__init__(models)
        for model in self._models:
            for tool in tools:
                model._agent._register_tool(tool)

    def run(self, prompt: str):
        return super().ask(prompt)



from typing import List

class FewShotPrompt:

    def __init__(self, prompt: str, examples: List[str]):
        self._prompt = prompt
        self._examples = examples

    def _format(self) -> str:
        return self._prompt + "\n\n" + \
            "Use the information from the following examples:" + "\n" + \
            "\n".join(self._examples)



from os import environ
import sys
import os
from typing import Dict, Optional

_KEY_EXT = "_API_KEY"

class KeyManager:

    """
    Key management module for AI service providers.

    This module handles the loading and validation of API keys for various AI service
    providers (like OpenAI, Anthropic, etc.). It manages key retrieval from environment
    variables and ensures proper key configuration before API usage.

    The module read keys from a providers.keys file in the root directory.
    The file is a simple text file with the following format:
    PROVIDER_NAME=API_KEY

    For example:
    OPENAI=sk-proj-ABCDE12345
    DEEPSEEK=sk-proj-FGHIJ67890

    Keys are automatically loaded when you use a model, so you don't need to use this module directly unless you want to load keys from a different file or disable key loading.
    """

    _instance = None
    _keys = None
    # Avoid loading configuration at import time; use default and optionally override on first use
    _key_file_path = "providers.keys"
    _is_enabled = True

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(KeyManager, cls).__new__(cls)
        return cls._instance

    def set_key_file(self, key_file_path: str):
        """
        Set the path to the keys file.

        This method allows you to specify a custom path for the keys file.
        By default, the file is expected to be named providers.keys and located in the root directory of the project.

        Parameters
        ----------
        key_file_path : str
            The path to the keys file
        """
        self._key_file_path = key_file_path

    def enabled(self, enable: bool):
        
        """
        Enable or disable key loading.

        This method allows you to toggle the key loading mechanism on and off.
        When disabled, the KeyManager will not attempt to load keys from the
        providers.keys file.

        Parameters
        ----------
        enable : bool
            True to enable key loading, False to disable
        """
        self._is_enabled = enable

    def load_key(self, provider: str):
        
        """
        Load and validate API key for a specific provider.

        This function checks for the presence of the appropriate API key in
        environment variables and validates its format. If the key is not found
        or is invalid, it raises an exception.

        Parameters
        ----------
        provider : str
            The name of the provider (e.g., 'openai', 'anthropic')

        Raises
        ------
        ValueError
            If the API key is not found in environment variables
        ValueError
            If the API key format is invalid
        
        This method is automatically called when you use a model, so you don't really need to use this method directly.
        """
        
        if not self._is_enabled:
            return

        key_name = provider.upper() + _KEY_EXT

        if self._keys is None:
            # Optionally pick key file path from configuration lazily on first use
            try:
                from monoai.conf import Conf  # local import to avoid eager YAML load
                cfg_path = Conf()["keysfile_path"]
                if cfg_path:
                    self._key_file_path = cfg_path
            except Exception:
                # Fallback to default path if configuration is unavailable
                pass
            self._load_keys_from_file()            

        key = self._keys.get(key_name)
        if key is None:
            raise ValueError(f"Key for {provider} not found")
        
        environ[key_name] = key


    def _load_keys_from_file(self):

        try:
            with open(self._key_file_path, 'r') as f:
                self._keys = {}
                for line in f:
                    line = line.strip()
                    if not line or line.startswith('#'):
                        continue
                    
                    if '=' in line:
                        provider, key = line.split('=', 1)
                        provider, key = provider.strip().upper(), key.strip().strip('"')
                        
                        if not provider.endswith(_KEY_EXT):
                            provider += _KEY_EXT
                            
                        self._keys[provider] = key
                    else:
                        print(f"Warning: Invalid line format in providers.keys: {line}")
        except FileNotFoundError:
            raise FileNotFoundError("providers.keys file not found. Please create it with your API keys.")



def load_key(provider: str):

    key_name = provider.upper() + _KEY_EXT

    if key_name in environ:
        return
    
    if 'google.colab' in sys.modules:
        _load_keys_from_colab(key_name)
    else:
        KeyManager().load_key(provider)


def _load_keys_from_colab(key_name: str):
    from google.colab import userdata
    environ[key_name] = userdata.get(key_name)



def check_key(provider: str) -> bool:
    """
    Check if an API key is available for a specific provider.

    Parameters
    ----------
    provider : str
        The name of the provider (e.g., 'openai', 'anthropic')

    Returns
    -------
    bool
        True if a key is available and valid, False otherwise

    Examples
    --------
    Check key availability:
        >>> if check_key("openai"):
        ...     print("OpenAI key is available")
        ... else:
        ...     print("OpenAI key is not available")
    """
    key = _get_key(provider)
    return bool(key and validate_key(provider, key))

def validate_key(provider: str, key: str) -> bool:
    """
    Validate the format of an API key for a specific provider.

    Parameters
    ----------
    provider : str
        The name of the provider (e.g., 'openai', 'anthropic')
    key : str
        The API key to validate

    Returns
    -------
    bool
        True if the key format is valid, False otherwise

    Examples
    --------
    Validate OpenAI key:
        >>> is_valid = validate_key("openai", "sk-...")
        >>> print(f"Key is {'valid' if is_valid else 'invalid'}")

    Validate Anthropic key:
        >>> is_valid = validate_key("anthropic", "sk-ant-...")
        >>> print(f"Key is {'valid' if is_valid else 'invalid'}")
    """
    if not key:
        return False
    
    validation_rules = {
        "openai": lambda k: k.startswith("sk-"),
        "anthropic": lambda k: k.startswith("sk-ant-"),
        # Add more provider validation rules as needed
    }
    
    validator = validation_rules.get(provider.lower())
    return bool(validator and validator(key))

def _get_key(provider: str) -> Optional[str]:
    """
    Get API key from environment variables for a specific provider.

    This is an internal function that handles the retrieval of API keys
    from environment variables using a standardized naming convention.

    Parameters
    ----------
    provider : str
        The name of the provider (e.g., 'openai', 'anthropic')

    Returns
    -------
    Optional[str]
        The API key if found, None otherwise

    Notes
    -----
    The function uses the following naming convention for environment variables:
    - OpenAI: OPENAI_API_KEY
    - Anthropic: ANTHROPIC_API_KEY
    - Other providers: PROVIDER_API_KEY (uppercase)
    """
    env_var = f"{provider.upper()}_API_KEY"
    return os.getenv(env_var)

# Optional: Provider-specific key validation patterns
KEY_PATTERNS: Dict[str, str] = {
    "openai": r"^sk-[a-zA-Z0-9]{32,}$",
    "anthropic": r"^sk-ant-[a-zA-Z0-9]{32,}$",
    # Add more patterns as needed
}

if __name__ == "__main__":
    load_key("openai")
    print(environ["OPENAI_API_KEY"])
    load_key("deepseek")
    print(environ["DEEPSEEK_API_KEY"])




from litellm import embedding
from monoai.keys.keys_manager import load_key
from monoai.rag.vectordb import ChromaVectorDB
from typing import List, Dict, Any, Optional

class RAG:
    """
    Retrieval-Augmented Generation (RAG) system for semantic search and document retrieval.
    
    This class provides a high-level interface for performing semantic search queries
    against a vector database. It supports multiple vector database backends and
    embedding providers for flexible deployment scenarios.
    
    The RAG system works by:
    1. Converting text queries into vector embeddings
    2. Searching the vector database for similar document embeddings
    3. Returning the most relevant documents based on semantic similarity
    
    Attributes:
        _vectorizer (str): The embedding model used for vectorization
        _db (str): Name of the vector database
        _vector_db (ChromaVectorDB): The vector database backend
    
    Examples:
    --------
     Basic usage with default settings:
        
    ```python
    # Initialize RAG with a database name
    rag = RAG(database="my_documents")
        
    # Perform a semantic search
    results = rag.query("What is machine learning?", k=5)
    ```
        
    Using with specific embedding provider:
        
    ```python
    # Initialize with OpenAI embeddings
    rag = RAG(
        database="my_documents",
        provider="openai",
        vectorizer="text-embedding-ada-002"
    )
        
    # Search for relevant documents
    results = rag.query("Explain neural networks", k=10)
    ```
        
    Working with different vector databases:
        
    ```python
    # Currently supports ChromaDB
    rag = RAG(
        database="my_collection",
        vector_db="chroma",
        provider="openai",
        vectorizer="text-embedding-ada-002"
    )
    ```

    Add RAG to a model, so that the model can use the RAG automatically to answer questions:
    ```python
    model = Model(provider="openai", model="gpt-4o-mini")
    model._add_rag(RAG(database="my_documents", vector_db="chroma"))
    ```

    """

    def __init__(self, 
                database: str,
                 provider: Optional[str] = None,
                 vectorizer: Optional[str] = None, 
                 vector_db: str = "chroma"):
        """
        Initialize the RAG system.
        
        Parameters:
        -----------
        database : str
            Name of the vector database/collection to use for storage and retrieval.
            This will be created if it doesn't exist.
            
        provider : str, optional
            The embedding provider to use (e.g., "openai", "anthropic", "cohere").
            If provided, the corresponding API key will be loaded automatically.
            If None, the system will use default embedding settings.
            
        vectorizer : str, optional
            The specific embedding model to use for vectorization.
            Examples: "text-embedding-ada-002", "text-embedding-3-small", "embed-english-v3.0"
            If None, the provider's default model will be used.
            
        vector_db : str, default="chroma"
            The vector database backend to use. Currently supports:
            - "chroma": ChromaDB (default, recommended for most use cases)
            
        Raises:
        -------
        ValueError
            If an unsupported vector database is specified.
            
        Examples:
        ---------
        ```python
        # Minimal initialization
        rag = RAG("my_documents")
        
        # With OpenAI embeddings
        rag = RAG(
            database="research_papers",
            provider="openai",
            vectorizer="text-embedding-ada-002"
        )
        
        # With Anthropic embeddings
        rag = RAG(
            database="articles",
            provider="anthropic",
            vectorizer="text-embedding-3-small"
        )
        ```
        """
        if provider:
            load_key(provider)
            
        self._vectorizer = vectorizer
        self._db = database
        
        if vector_db == "chroma":
            self._vector_db = ChromaVectorDB(
                name=database, 
                vectorizer_provider=provider, 
                vectorizer_model=vectorizer
            )
        else:
            raise ValueError(f"Vector database '{vector_db}' not supported. Currently only 'chroma' is supported.")
        

    def query(self, query: str, k: int = 10) -> Dict[str, Any]:
        """
        Perform a semantic search query against the vector database.
        
        This method converts the input query into a vector embedding and searches
        the database for the most semantically similar documents.
        
        Parameters:
        -----------
        query : str
            The text query to search for. This will be converted to a vector
            embedding and used to find similar documents.
            
        k : int, default=10
            The number of most relevant documents to return. Higher values
            return more results but may include less relevant documents.
            
        Returns:
        --------
        Dict[str, Any]
            A dictionary containing the search results with the following structure:
            {
                'ids': List[List[str]] - Document IDs of the retrieved documents,
                'documents': List[List[str]] - The actual document content,
                'metadatas': List[List[Dict]] - Metadata for each document,
                'distances': List[List[float]] - Similarity scores (lower = more similar)
            }
            
        Examples:
        ---------
        ```python
        # Basic query
        results = rag.query("What is artificial intelligence?")
        
        # Query with more results
        results = rag.query("Machine learning algorithms", k=20)
        
        # Accessing results
        for i, (doc_id, document, metadata, distance) in enumerate(zip(
            results['ids'][0], 
            results['documents'][0], 
            results['metadatas'][0], 
            results['distances'][0]
        )):
            print(f"Result {i+1}:")
            print(f"  ID: {doc_id}")
            print(f"  Content: {document[:100]}...")
            print(f"  Similarity: {1 - distance:.3f}")
            print(f"  Metadata: {metadata}")
            print()
        ```
        
        Notes:
        ------
        - The query is automatically converted to lowercase and processed
        - Results are returned in order of relevance (most similar first)
        - Distance scores are cosine distances (0 = identical, 2 = completely opposite)
        - If fewer than k documents exist in the database, all available documents are returned
        """
        return self._vector_db.query(query, k)








from monoai.models import Model
from monoai.prompts import Prompt, PromptChain

TEST_PROMPT = "This is a test"
TEST_PROMPT_FOLDER = "monoai/tests/prompts"
TEST_MODEL = "gpt-4o-nano"

def test_base_model():
    assert "response" in Model(provider="openai", model=TEST_MODEL).ask(TEST_PROMPT)

def test_base_model_with_prompt():
    prompt = Prompt(prompt_id=f"{TEST_PROMPT_FOLDER}/base")
    assert "response" in Model(provider="openai", model=TEST_MODEL).ask(prompt=prompt)

def test_model_with_system_prompt():
    system_prompt = "You are a helpful assistant"
    model = Model(provider="openai", model=TEST_MODEL, system_prompt=system_prompt)
    response = model.ask(TEST_PROMPT)
    assert "response" in response
    assert response["model"]["provider"] == "openai"
    assert response["model"]["name"] == TEST_MODEL

def test_model_with_token_counting():
    model = Model(provider="openai", model=TEST_MODEL, count_tokens=True)
    response = model.ask(TEST_PROMPT)
    assert "response" in response

def test_model_with_cost_counting():
    model = Model(provider="openai", model=TEST_MODEL, count_tokens=True, count_cost=True)
    response = model.ask(TEST_PROMPT)
    assert "response" in response

def test_model_with_prompt_chain():
    chain = PromptChain(prompts=[
        "What is 2+2?",
        "Based on the previous answer, what is 2+2+2?"
    ])
    model = Model(provider="openai", model=TEST_MODEL)
    response = model.ask(chain)
    assert "response" in response
    assert isinstance(response["response"], str)
    assert len(response["response"]) > 0

def test_model_with_prompt_variables():
    prompt = Prompt(
        prompt_id=f"{TEST_PROMPT_FOLDER}/with_variables",
        prompt_data={"country": "Italy"}
    )
    model = Model(provider="openai", model=TEST_MODEL)
    response = model.ask(prompt)
    assert "response" in response
    assert isinstance(response["response"], str)
    assert len(response["response"]) > 0
    assert "Rome" in response["response"]

def test_model_with_formatted_prompt():
    test_text = "The quick brown fox jumps over the lazy dog. This is a test text for summarization."
    prompt = Prompt(
        prompt_id=f"{TEST_PROMPT_FOLDER}/with_formatting",
        prompt_data={"text": test_text}
    )
    model = Model(provider="openai", model=TEST_MODEL)
    response = model.ask(prompt)
    assert "response" in response
    assert isinstance(response["response"], str)
    assert len(response["response"]) > 0
    assert "-" in response["response"]  # Check for bullet points


def test_model_with_type():
    prompt = Prompt(
        prompt="2+2=",
        response_type="int"
    )
    model = Model(provider="openai", model=TEST_MODEL)
    response = model.ask(prompt)
    assert "response" in response
    assert isinstance(response["response"], int)
    assert response["response"] == 4


def test_model_with_prompt_type():
    prompt = Prompt(prompt_id=f"{TEST_PROMPT_FOLDER}/with_type")
    model = Model(provider="openai", model=TEST_MODEL)
    response = model.ask(prompt)
    assert "response" in response
    assert isinstance(response["response"], int)
    assert response["response"] == 4




def search_web(query: str, engine: str = "duckduckgo", max_results: int = 5, exclude_domains: list[str] = None):
    """Search the web using the specified search engine.
    
    Args:
        query: The query to search for
        engine: The search engine to use (duckduckgo or tavily). Default is duckduckgo
        max_results: The maximum number of results to return. Default is 5
        exclude_domains: The domains to exclude from the search. Default is None
    
    Returns:
        A dictionary containing:
            data: The search results as a list of dictionaries
            text: The results merged into a single string
            
    Raises:
        ValueError: If an invalid engine is specified
        
    Examples:
        Basic usage with DuckDuckGo:
        ```python
        result = search_web("What is the capital of France?")
        print(result["text"])  # print the result merged into a single string
        print(result["data"])  # print the result as a list of dictionaries
        ```
        
        Using Tavily with custom parameters:
        ```python
        result = search_web("Python programming", engine="tavily", max_results=10)
        print(result["data"])  # print the search results
        ```
    """
    
    if engine == "duckduckgo":
        search_engine = _DuckDuckGoSearch(max_results, exclude_domains)
    elif engine == "tavily":
        search_engine = _TavilySearch(max_results, exclude_domains)
    else:
        raise ValueError(f"Invalid engine: {engine} (must be 'duckduckgo' or 'tavily')")
    
    response, text_response = search_engine.search(query)
    return {"data": response, "text": text_response}

from duckduckgo_search import DDGS

class _BaseSearch():

    def __init__(self, max_results: int = 5, exclude_domains: list[str] = None):
        self._max_results = max_results
        self._exclude_domains = exclude_domains

    def search(self, query: str):
        pass

    def _post_process(self, response: list[dict], title_key: str = "title", text_key: str = "body", url_key: str = "url"):
        response = [{"title": item[title_key], "text": item[text_key], "url": item[url_key]} for item in response]
        text_response = "\n\n".join([item["title"] + "\n\n" + item["text"] for item in response])
        return response, text_response

class _DuckDuckGoSearch(_BaseSearch):

    def __init__(self, max_results: int = 5, exclude_domains: list[str] = None):
        super().__init__(max_results, exclude_domains)
        self._client = DDGS()

    def search(self, query: str):
        response = self._client.text(query, max_results=self._max_results)
        return self._post_process(response, title_key="title", text_key="body", url_key="href")
    

from tavily import TavilyClient
from monoai.keys.keys_manager import load_key

class _TavilySearch(_BaseSearch):

    def __init__(self, max_results: int = 5, exclude_domains: list[str] = None):
        super().__init__(max_results, exclude_domains)
        load_key("tavily")
        self._client = TavilyClient()

    def search(self, query: str):
        response = self._client.search(query, max_results=self._max_results, exclude_domains=self._exclude_domains)
        response = response["results"]
        return self._post_process(response, title_key="title", text_key="content", url_key="url")



import json
import inspect
from typing import Any, Dict, List, Optional
from ..prompts import Prompt


class _FunctionCallingMixin:
    """Mixin per gestire chiamate di funzioni con tool calls OpenAI.
    
    Questo mixin fornisce metodi per gestire le chiamate di tool in formato
    OpenAI function calling, convertendo le risposte in messaggi standardizzati.
    """
    
    def _call_tool(self, tool_call: Any) -> Dict[str, str]:
        """Esegue una chiamata di tool e restituisce la risposta formattata.
        
        Parameters
        ----------
        tool_call : Any
            Oggetto tool_call contenente informazioni sulla funzione da chiamare.
            Deve avere attributi `function.name`, `function.arguments` e `id`.
        
        Returns
        -------
        Dict[str, str]
            Dizionario con la risposta del tool formattata per i messaggi:
            - tool_call_id: ID della chiamata tool
            - role: Ruolo del messaggio (sempre "tool")
            - name: Nome della funzione chiamata
            - content: Risposta del tool come stringa
        """
        function_name = tool_call.function.name
        function_to_call = self._tools[function_name]
        function_args = json.loads(tool_call.function.arguments)                
        function_response = str(function_to_call(**function_args))
        return {
            "tool_call_id": tool_call.id,
            "role": "tool",
            "name": function_name,
            "content": function_response,
        }


class _ReactMixin:
    """Mixin per gestire chiamate di tool in formato React/JSON.
    
    Questo mixin fornisce metodi per gestire le chiamate di tool in formato
    JSON strutturato, tipico degli approcci React-style per gli agenti AI.
    """
    
    def _encode_tool(self, func: Any) -> str:
        """Codifica una funzione in formato stringa descrittivo.
        
        Parameters
        ----------
        func : Any
            Funzione da codificare. Deve avere attributi `__name__` e `__doc__`.
        
        Returns
        -------
        str
            Stringa descrittiva della funzione nel formato:
            "nome_funzione(signature): documentazione"
        """
        sig = inspect.signature(func)
        doc = inspect.getdoc(func)
        encoded = func.__name__ + str(sig) + ": " + doc
        encoded = encoded.replace("\n", " ")
        return encoded
    
    def _call_tool(self, tool_call: Dict[str, Any]) -> Any:
        """Esegue una chiamata di tool in formato React.
        
        Parameters
        ----------
        tool_call : Dict[str, Any]
            Dizionario con le informazioni del tool:
            - name: Nome del tool da chiamare
            - arguments: Argomenti del tool come dizionario
        
        Returns
        -------
        Any
            Risultato dell'esecuzione del tool
        """
        tool = self._tools[tool_call["name"]]
        kwargs = list(tool_call["arguments"].values())
        return tool(*kwargs)


class _AgenticLoop:
    """Classe base per tutti i loop agentici.
    
    Questa classe fornisce la funzionalità di base per tutti gli agenti AI,
    inclusa la gestione degli strumenti, la creazione di messaggi e l'esecuzione
    del modello. È progettata per essere estesa da classi specifiche che
    implementano diversi approcci agentici.
    
    Attributes
    ----------
    _model : Any
        Modello AI da utilizzare per l'esecuzione
    _agentic_prompt : str
        Prompt personalizzato per l'agente (opzionale)
    _debug : bool
        Flag per abilitare la stampa di debug
    _max_iter : Optional[int]
        Numero massimo di iterazioni consentite
    _tools : Dict[str, Any]
        Dizionario degli strumenti disponibili, mappati per nome
    """
    
    def __init__(self, model: Any, agentic_prompt: str=None, debug: bool=False, max_iter: Optional[int]=None) -> None:
        """Inizializza l'agente con il modello e gli strumenti.
        
        Parameters
        ----------
        model : Any
            Modello AI da utilizzare per l'esecuzione
        agentic_prompt : str
            Prompt personalizzato per l'agente (None per usare quello di default)
        debug : bool
            Flag per abilitare la stampa di debug
        max_iter : Optional[int]
            Numero massimo di iterazioni consentite (None per illimitato)
        """
        self._model = model
        self._agentic_prompt = agentic_prompt
        self._debug = debug
        self._max_iter = max_iter
        self._tools = {}


    def register_tools(self, tools: List[Any]) -> None:
        for tool in tools:
            self._tools[tool.__name__] = tool

    def _get_tools(self) -> str:
        """Genera la stringa descrittiva degli strumenti disponibili.
        
        Returns
        -------
        str
            Stringa formattata con la descrizione di tutti gli strumenti disponibili,
            uno per riga con prefisso " - "
        """
        if not self._tools:
            return ""
        
        tools = []
        for tool_name, tool_func in self._tools.items():
            tools.append(f" - {self._encode_tool(tool_func)}")
        return "\n".join(tools)

    def _get_base_messages(self, agent_type: str, query: str) -> List[Dict[str, Any]]:
        """Genera i messaggi base per l'agente specifico.
        
        Parameters
        ----------
        agent_type : str
            Tipo di agente per determinare il prompt da utilizzare
        query : str
            Query dell'utente da includere nel prompt
        
        Returns
        -------
        List[Dict[str, Any]]
            Lista dei messaggi base per l'agente, inclusi il prompt e la query
        """
        tools = self._get_tools()
        prompt_id = (f"monoai/agents/prompts/{agent_type}.prompt" 
                    if self._agentic_prompt is None else self._agentic_prompt)
        
        prompt = Prompt(
            prompt_id=prompt_id,
            prompt_data={"query": query, "available_tools": tools}
        )
        
        return [prompt.as_dict()]

    def _debug_print(self, content: str) -> None:
        """Stampa debug se abilitato.
        
        Parameters
        ----------
        content : str
            Contenuto da stampare in modalità debug
        """
        if self._debug:
            print(content)
            print("-------")

    def _execute_model_step(self, messages: List[Dict[str, Any]]) -> Dict[str, Any]:
        """Esegue un passo del modello e restituisce la risposta.
        
        Parameters
        ----------
        messages : List[Dict[str, Any]]
            Lista dei messaggi da inviare al modello
        
        Returns
        -------
        Dict[str, Any]
            Risposta del modello nel formato standard OpenAI
        """
        
        resp = self._model._execute(messages)
        return resp["choices"][0]["message"]

    def _create_base_response(self, query: str) -> Dict[str, Any]:
        """Crea la struttura base della risposta.
        
        Parameters
        ----------
        query : str
            Query originale dell'utente
        
        Returns
        -------
        Dict[str, Any]
            Dizionario con la struttura base della risposta:
            - prompt: Query originale
            - iterations: Lista vuota per le iterazioni
        """
        return {"prompt": query, "iterations": []}

    def _handle_final_answer(self, iteration: Dict[str, Any], response: Dict[str, Any]) -> bool:
        """Gestisce una risposta finale, restituisce True se è la fine.
        
        Parameters
        ----------
        iteration : Dict[str, Any]
            Iterazione corrente contenente potenzialmente una risposta finale
        response : Dict[str, Any]
            Dizionario della risposta da aggiornare
        
        Returns
        -------
        bool
            True se è stata trovata una risposta finale, False altrimenti
        
        Notes
        -----
        Questo metodo modifica direttamente l'oggetto response passato come parametro.
        """
        if "final_answer" in iteration:
            response["iterations"].append(iteration)
            response["response"] = iteration["final_answer"]
            return True
        return False

    def _handle_tool_action(self, iteration: Dict[str, Any], response: Dict[str, Any], messages: List[Dict[str, Any]]) -> None:
        """Gestisce un'azione di tool.
        
        Parameters
        ----------
        iteration : Dict[str, Any]
            Iterazione corrente contenente l'azione del tool
        response : Dict[str, Any]
            Dizionario della risposta da aggiornare
        messages : List[Dict[str, Any]]
            Lista dei messaggi da aggiornare con l'osservazione
        
        Notes
        -----
        Questo metodo modifica direttamente gli oggetti response e messages passati come parametri.
        """
        if "action" in iteration and iteration["action"].get("name"):
            tool_call = iteration["action"]
            tool_result = self._call_tool(tool_call)
            iteration["observation"] = tool_result
            response["iterations"].append(iteration)
            
            msg = json.dumps({"observation": tool_result})
            messages.append({"type": "user", "content": msg})

    def _handle_default(self, iteration: Dict[str, Any], response: Dict[str, Any], messages: List[Dict[str, Any]]) -> None:
        """Gestisce il caso default per le iterazioni non gestite.
        
        Parameters
        ----------
        iteration : Dict[str, Any]
            Iterazione corrente da gestire
        response : Dict[str, Any]
            Dizionario della risposta da aggiornare
        messages : List[Dict[str, Any]]
            Lista dei messaggi da aggiornare
        
        Notes
        -----
        Questo metodo modifica direttamente gli oggetti response e messages passati come parametri.
        """
        response["iterations"].append(iteration)
        messages.append({"type": "user", "content": json.dumps(iteration)})

    def start(self, query: str) -> Dict[str, Any]:
        """Metodo astratto per avviare il loop agentico.
        
        Parameters
        ----------
        query : str
            Query dell'utente da processare
        
        Returns
        -------
        Dict[str, Any]
            Risposta dell'agente con le iterazioni e il risultato finale
        
        Raises
        ------
        NotImplementedError
            Questo metodo deve essere implementato dalle classi figlie
        """
        raise NotImplementedError


class FunctionCallingAgenticLoop(_AgenticLoop, _FunctionCallingMixin):
    """Agente che utilizza function calling OpenAI.
    
    Questo agente implementa un loop che utilizza il sistema di function calling
    nativo di OpenAI, permettendo al modello di chiamare direttamente le funzioni
    disponibili senza parsing manuale delle risposte.
    
    Attributes
    ----------
    _model : Any
        Modello OpenAI con supporto per function calling
    _tools : Dict[str, Any]
        Strumenti disponibili per l'agente
    """
    
    def start(self, query: str) -> Dict[str, Any]:
        """Avvia il loop agentico utilizzando function calling.
        
        Parameters
        ----------
        query : str
            Query dell'utente da processare
        
        Returns
        -------
        Dict[str, Any]
            Risposta dell'agente con:
            - prompt: Query originale
            - iterations: Lista delle chiamate di tool eseguite
            - response: Risposta finale del modello
        """
        self._model._add_tools(list(self._tools.values()))
        messages = [{"type": "user", "content": query}]
        response = self._create_base_response(query)
        
        while True:
            resp = self._execute_model_step(messages)
            messages.append(resp)
            content = resp["content"]

            self._debug_print(content)

            if content is not None:
                response["response"] = content
                break
            
            if resp.get("tool_calls"):
                for tool_call in resp["tool_calls"]:
                    tool_result = self._call_tool(tool_call)
                    response["iterations"].append({
                        "name": tool_call.function.name,
                        "arguments": tool_call.function.arguments,
                        "result": tool_result["content"]
                    })
                    messages.append(tool_result)
        
        return response


class _BaseReactLoop(_AgenticLoop, _ReactMixin):
    """Classe base per tutti gli agenti React-style.
    
    Questa classe implementa il loop standard per gli agenti che utilizzano
    un approccio React-style, dove il modello produce risposte JSON strutturate
    che vengono parse e gestite iterativamente.
    
    Attributes
    ----------
    _max_iter : Optional[int]
        Numero massimo di iterazioni consentite
    """
    
    def _run_react_loop(self, query: str, agent_type: str, 
                        custom_handlers: Optional[Dict[str, callable]] = None) -> Dict[str, Any]:
        """Esegue il loop React standard.
        
        Parameters
        ----------
        query : str
            Query dell'utente da processare
        agent_type : str
            Tipo di agente per determinare il prompt da utilizzare
        custom_handlers : Optional[Dict[str, callable]], optional
            Dizionario di handler personalizzati per gestire tipi specifici
            di iterazioni. Le chiavi sono i campi nell'iterazione, i valori
            sono funzioni che gestiscono quelle iterazioni.
        
        Returns
        -------
        Dict[str, Any]
            Risposta dell'agente con:
            - prompt: Query originale
            - iterations: Lista delle iterazioni processate
            - response: Risposta finale (se presente)
        
        Notes
        -----
        Questo metodo gestisce automaticamente:
        - Risposte finali (final_answer)
        - Azioni di tool (action)
        - Casi personalizzati tramite custom_handlers
        - Casi default per iterazioni non gestite
        - Gestione degli errori JSON
        """
        messages = self._get_base_messages(agent_type, query)
        current_iter = 0
        response = self._create_base_response(query)
        
        # Handler personalizzati per casi speciali
        custom_handlers = custom_handlers or {}

        while True:
            if self._max_iter is not None and current_iter >= self._max_iter:
                break
            
            resp = self._execute_model_step(messages)
            messages.append(resp)
            content = resp["content"]

            self._debug_print(content)

            if content is not None:
                try:
                    iteration = json.loads(content)
                    
                    # Gestione risposta finale
                    if self._handle_final_answer(iteration, response):
                        break
                    
                    # Gestione azioni di tool
                    if "action" in iteration:
                        self._handle_tool_action(iteration, response, messages)
                        continue
                    
                    # Gestione casi personalizzati
                    handled = False
                    for key, handler in custom_handlers.items():
                        if key in iteration and iteration[key] is not None:
                            handler(iteration, response, messages)
                            handled = True
                            break
                    
                    if not handled:
                        self._handle_default(iteration, response, messages)
                        
                except json.JSONDecodeError:
                    # Se non è JSON valido, aggiungi come messaggio utente
                    messages.append({"type": "user", "content": content})

            current_iter += 1

        return response


class ReactAgenticLoop(_BaseReactLoop):
    """Agente React standard.
    
    Questo agente implementa il pattern React standard, dove il modello
    produce risposte JSON che vengono parse e gestite iterativamente.
    """
    
    def start(self, query: str) -> Dict[str, Any]:
        """Avvia il loop agentico React.
        
        Parameters
        ----------
        query : str
            Query dell'utente da processare
        
        Returns
        -------
        Dict[str, Any]
            Risposta dell'agente processata tramite il loop React
        """
        return self._run_react_loop(query, "react")


class ReactWithFCAgenticLoop(_AgenticLoop, _FunctionCallingMixin):
    """Agente che combina React e Function Calling.
    
    Questo agente combina l'approccio React con il function calling nativo
    di OpenAI, permettendo una gestione ibrida delle chiamate di tool.
    """
    
    def start(self, query: str) -> Dict[str, Any]:
        """Avvia il loop agentico ibrido React + Function Calling.
        
        Parameters
        ----------
        query : str
            Query dell'utente da processare
        
        Returns
        -------
        Dict[str, Any]
            Risposta dell'agente (da implementare)
        
        Notes
        -----
        TODO: Implementare combinazione di React e Function Calling
        """
        # TODO: Implementare combinazione di React e Function Calling
        pass


class ProgrammaticAgenticLoop(_BaseReactLoop):
    """Agente programmatico.
    
    Questo agente implementa un approccio programmatico dove il modello
    produce codice o istruzioni strutturate che vengono eseguite.
    """
    
    def start(self, query: str) -> Dict[str, Any]:
        """Avvia il loop agentico programmatico.
        
        Parameters
        ----------
        query : str
            Query dell'utente da processare
        
        Returns
        -------
        Dict[str, Any]
            Risposta dell'agente processata tramite il loop programmatico
        """
        return self._run_react_loop(query, "programmatic")


class PlanAndExecuteAgenticLoop(_BaseReactLoop):
    """Agente plan-and-execute.
    
    Questo agente implementa il pattern plan-and-execute, dove il modello
    prima pianifica le azioni e poi le esegue sequenzialmente.
    """
    
    def start(self, query: str) -> Dict[str, Any]:
        """Avvia il loop agentico plan-and-execute.
        
        Parameters
        ----------
        query : str
            Query dell'utente da processare
        
        Returns
        -------
        Dict[str, Any]
            Risposta dell'agente processata tramite il loop plan-and-execute
        """
        return self._run_react_loop(query, "plan_and_execute")


class ReflexionAgenticLoop(_BaseReactLoop):
    """Agente con riflessione.
    
    Questo agente implementa il pattern reflexion, dove il modello
    riflette sulle proprie azioni e decisioni per migliorare le performance.
    """
    
    def start(self, query: str) -> Dict[str, Any]:
        """Avvia il loop agentico con riflessione.
        
        Parameters
        ----------
        query : str
            Query dell'utente da processare
        
        Returns
        -------
        Dict[str, Any]
            Risposta dell'agente processata tramite il loop reflexion
        """
        return self._run_react_loop(query, "reflexion")


class SelfAskAgenticLoop(_BaseReactLoop):
    """Agente self-ask.
    
    Questo agente implementa il pattern self-ask, dove il modello
    si pone domande a se stesso per guidare il processo di ragionamento.
    """
    
    def start(self, query: str) -> Dict[str, Any]:
        """Avvia il loop agentico self-ask.
        
        Parameters
        ----------
        query : str
            Query dell'utente da processare
        
        Returns
        -------
        Dict[str, Any]
            Risposta dell'agente processata tramite il loop self-ask
        """
        return self._run_react_loop(query, "self_ask")


class SelfAskWithSearchLoop(_BaseReactLoop):
    """Agente self-ask con ricerca web.
    
    Questo agente estende il pattern self-ask con la capacità di
    eseguire ricerche web per ottenere informazioni aggiuntive.
    
    Attributes
    ----------
    _handle_search_query : callable
        Metodo per gestire le query di ricerca web
    """
    
    def _handle_search_query(self, iteration: Dict[str, Any], response: Dict[str, Any], messages: List[Dict[str, Any]]) -> None:
        """Gestisce le query di ricerca web.
        
        Parameters
        ----------
        iteration : Dict[str, Any]
            Iterazione corrente contenente la query di ricerca
        response : Dict[str, Any]
            Dizionario della risposta da aggiornare
        messages : List[Dict[str, Any]]
            Lista dei messaggi da aggiornare con i risultati della ricerca
        
        Notes
        -----
        Questo metodo modifica direttamente gli oggetti response e messages passati come parametri.
        Utilizza il motore di ricerca Tavily per eseguire le ricerche web.
        """
        from ..tools.websearch import search_web
        
        query = iteration["search_query"]
        result = search_web(query, engine="tavily")["text"]
        iteration["search_result"] = result
        
        msg = json.dumps({"query_results": result})
        messages.append({"type": "user", "content": msg})
        response["iterations"].append(iteration)
    
    def start(self, query: str) -> Dict[str, Any]:
        """Avvia il loop agentico self-ask con ricerca web.
        
        Parameters
        ----------
        query : str
            Query dell'utente da processare
        
        Returns
        -------
        Dict[str, Any]
            Risposta dell'agente processata tramite il loop self-ask con ricerca web
        
        Notes
        -----
        Questo agente utilizza un handler personalizzato per gestire le query di ricerca
        web, permettendo al modello di ottenere informazioni aggiornate durante il processo.
        """
        custom_handlers = {"search_query": self._handle_search_query}
        return self._run_react_loop(query, "self_ask_with_search", custom_handlers)



from litellm import speech
from ..keys.keys_manager import load_key
import base64
import io
from elevenlabs.client import ElevenLabs
import os

class VoiceModel:
    
    def __init__(self, provider: str, model: str, voice: str):
        load_key(provider)
        self._provider = provider
        self._model = model
        self._voice = voice
        
        if self._provider == "elevenlabs":
            self._client = ElevenLabs(api_key=os.getenv("ELEVENLABS_API_KEY"))
        
    def speak(self, text: str, min_chars_per_sentence=100, return_type="audio.mp3"):
        
        audio_groups = self._generate_audio_groups(text, min_chars_per_sentence)
        audio_chunks = []
        
        for group in audio_groups:
            
            response = self._generate(group)
            
            audio_chunks.append(response)
        
        if return_type.endswith(".mp3") or return_type.endswith(".wav"):
            combined_audio = self._combine_bytes_chunks(audio_chunks)
            with open(return_type, 'wb') as f:
                f.write(combined_audio)
            return return_type
        elif return_type == "base64":
            combined_audio = self._combine_bytes_chunks(audio_chunks)
            return base64.b64encode(combined_audio).decode('utf-8')
        elif return_type == "bytes":
            return self._combine_bytes_chunks(audio_chunks)
        else:
            raise ValueError(f"Invalid return type: {return_type}")
    
    def _combine_bytes_chunks(self, audio_chunks: list) -> bytes:
        """Unisce i chunk audio in un unico oggetto bytes"""
        combined = b""
        for chunk in audio_chunks:
            combined += chunk
        return combined
    
    def _generate_audio_groups(self, text: str, min_chars_per_sentence: int):
        """Genera gruppi di sentences ottimizzati per lunghezza"""
        sentences = [s.strip() for s in text.split(".") if s.strip()]
        audio_groups = []
        i = 0
        
        while i < len(sentences):
            current_group = sentences[i]
            j = i + 1
            
            # Continua ad aggiungere sentences successive finché non superi min_chars_per_sentence
            while j < len(sentences):
                next_sentence = sentences[j]
                combined_length = len(current_group + ". " + next_sentence)
                
                # Se l'unione supera il limite, fermati
                if combined_length > min_chars_per_sentence:
                    break
                
                # Altrimenti, unisci la sentence successiva
                current_group += ". " + next_sentence
                j += 1
            
            audio_groups.append(current_group)
            # Passa alla prossima sentence non processata
            i = j
        
        return audio_groups
        
    def _generate(self, text):
        
        if self._provider == "elevenlabs":
            response = self._client.text_to_speech.convert(
                text=text,
                voice_id=self._voice,
                model_id=self._model,
                output_format="mp3_44100_128",
            )
            response_bytes = b""
            for r in response:
                response_bytes += r
            return response_bytes
        else:
            response = speech(
                model=self._provider+"/"+self._model,
                voice=self._voice,
                input=text,
            )
            return response.content
    
    async def stream(self, text: str, min_chars_per_sentence=100, return_type="audio.mp3"):
        
        audio_groups = self._generate_audio_groups(text, min_chars_per_sentence)
        
        for chunk_index, group in enumerate(audio_groups):
            # Genera audio per il gruppo di sentences
            response = self._generate(group)

            if return_type.endswith(".mp3") or return_type.endswith(".wav"):
                with open(return_type, 'wb') as f:
                    f.write(response)
                yield return_type
            elif return_type == "base64":
                yield base64.b64encode(response).decode('utf-8')
            elif return_type == "bytes":
                yield response
            else:
                raise ValueError(f"Invalid return type: {return_type}")



from litellm import completion, acompletion
from monoai.keys.keys_manager import load_key
from monoai.chat.history import *
from monoai.models import Model
from monoai.conf.conf import Conf
from monoai.prompts.prompt import Prompt
import os
import base64
import json
import logging
from typing import Union, Optional, AsyncGenerator, List, Dict, Any
from pathlib import Path

# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

class ChatError(Exception):
    """Custom exception for Chat-related errors."""
    pass

class Chat():
    """
    Chat class is responsible for handling the chat interface and messages history.
    
    Examples
    --------
    Basic usage:
    ```
    chat = Chat(provider="openai", model="gpt-4o-mini")
    response = chat.ask("2+2") # 4
    response = chat.ask("+2") # 6
    ```    

    With history:
    ```

    # Create a new chat with JSON history
    chat = Chat(provider="openai", model="gpt-4o-mini", history_type="json")
    print(chat.chat_id) # 8cc2bfa3-e9a0-4b82-b46e-3376cd220dd3
    response = chat.ask("Hello! I'm Giuseppe") # Hello!

    # Load a chat with JSON history
    chat = Chat(provider="openai", model="gpt-4o-mini", history_type="json", chat_id="8cc2bfa3-e9a0-4b82-b46e-3376cd220dd3")
    response = chat.ask("What's my name?") # Your name is Giuseppe

    # Create a new chat with in-memory dictionary history
    chat = Chat(provider="openai", model="gpt-4o-mini", history_type="dict")
    print(chat.chat_id) # 8cc2bfa3-e9a0-4b82-b46e-3376cd220dd3
    response = chat.ask("Hello! I'm Giuseppe") # Hello!
    ```

    With history summarizer:

    ```
    chat = Chat(provider="openai", 
                model="gpt-4o-mini", 
                history_type="json", 
                history_summarizer_provider="openai", 
                history_summarizer_model="gpt-4o-mini", 
                history_summarizer_max_tokens=100)
                
    response = chat.ask("Hello! I'm Giuseppe") # Hello!
    response = chat.ask("What's my name?") # Your name is Giuseppe
    ```
    """

    _HISTORY_MAP = {
        "json": JSONHistory,
        "sqlite": SQLiteHistory,
        "mongodb": MongoDBHistory,
        "dict": DictHistory
    }

    def __init__(self, 
                 model,
                 system_prompt: Optional[Union[Prompt, str]] = "You are an helpful assistant",
                 max_tokens: Optional[int] = None,
                 history_type: Union[str, BaseHistory] = "dict", 
                 history_last_n: Optional[int] = None,
                 history_path: Optional[str] = "histories",
                 history_summarizer_provider: Optional[str] = None, 
                 history_summarizer_model: Optional[str] = None,
                 history_summarizer_max_tokens: Optional[int] = None,
                 chat_id: Optional[str] = None) -> None:

        """
        Initialize a new Chat instance.

        Parameters
        ----------
        model : Model
            Model instance to use for the chat
        system_prompt : str | Prompt, optional
            System prompt or Prompt object
        max_tokens : int, optional
            Maximum number of tokens for each request
        history_type : str | BaseHistory, optional
            The type of history to use for the chat. Options: "json", "sqlite", "mongodb", "dict"
        history_last_n : int, optional
            The last n messages to keep in the history.
        history_path : str, optional
            The path to the history (not used for "dict" history type)
        history_summarizer_provider : str, optional
            The provider of the history summarizer.
        history_summarizer_model : str, optional
            The model of the history summarizer.
        history_summarizer_max_tokens : int, optional
            The maximum number of tokens for the history summarizer.
        chat_id : str, optional
            The id of the chat to load, if not provided a new chat will be created

        Raises
        ------
        ChatError
            If invalid parameters are provided or initialization fails
        """
        try:
            
            self._max_tokens = max_tokens
            self._model = model
            
            self._history_summarizer = None

            # Initialize history
            self._initialize_history(history_type, history_last_n, history_path)
            
            # Initialize history summarizer
            self._initialize_history_summarizer(
                history_summarizer_provider, 
                history_summarizer_model, 
                history_summarizer_max_tokens
            )

            # Process system prompt
            processed_system_prompt = self._process_system_prompt(system_prompt)

            # Initialize chat
            if chat_id is None:
                self.chat_id = self._history.new(processed_system_prompt)
            else:
                self.chat_id = chat_id
                
        except Exception as e:
            logger.error(f"Failed to initialize Chat: {e}")
            raise ChatError(f"Chat initialization failed: {e}")


    def _initialize_history(self, history_type: Union[str, BaseHistory], history_last_n: Optional[int], history_path: Optional[str]) -> None:
        """Initialize the history system."""
        try:
            if isinstance(history_type, str):
                if history_type == "dict":
                    # DictHistory doesn't need db_path
                    self._history = self._HISTORY_MAP[history_type](last_n=history_last_n)
                else:
                    self._history = self._HISTORY_MAP[history_type](last_n=history_last_n, path=history_path)
            else:
                self._history = history_type
        except Exception as e:
            raise ChatError(f"Failed to initialize history: {e}")

    def _initialize_history_summarizer(self, provider: Optional[str], model: Optional[str], max_tokens: Optional[int]) -> None:
        """Initialize the history summarizer."""
        if provider is not None and model is not None:
            try:
                self._history_summarizer = HistorySummarizer(
                    Model(provider=provider, model=model, max_tokens=max_tokens)
                )
            except Exception as e:
                logger.warning(f"Failed to initialize history summarizer: {e}")

    def _process_system_prompt(self, system_prompt: Optional[Union[Prompt, str]]) -> str:
        """Process and load the system prompt."""
        try:
            if system_prompt is None:
                return self._load_default_system_prompt()
            elif isinstance(system_prompt, str) and system_prompt.endswith(".prompt"):
                return self._load_prompt_file(system_prompt)
            elif isinstance(system_prompt, Prompt):
                return str(system_prompt)
            else:
                return system_prompt
        except Exception as e:
            logger.warning(f"Failed to process system prompt: {e}")
            return ""

    def _load_default_system_prompt(self) -> str:
        """Load the default system prompt from file."""
        try:
            prompt_path = Conf()["prompts_path"]
            system_prompt_path = Path(prompt_path) / "system.prompt"
            
            if system_prompt_path.exists():
                with open(system_prompt_path, "r", encoding="utf-8") as f:
                    return f.read()
            else:
                logger.info("Default system prompt file not found, using empty prompt")
                return ""
        except Exception as e:
            logger.warning(f"Failed to load default system prompt: {e}")
            return ""

    def _load_prompt_file(self, prompt_file: str) -> str:
        """Load a prompt from a file."""
        try:
            prompt_path = Conf()["prompts_path"]
            full_path = Path(prompt_path) / prompt_file
            
            if not full_path.exists():
                raise FileNotFoundError(f"Prompt file not found: {full_path}")
                
            with open(full_path, "r", encoding="utf-8") as f:
                return f.read()
        except Exception as e:
            raise ChatError(f"Failed to load prompt file {prompt_file}: {e}")

    def _process_file_attachment(self, prompt: str, file: Optional[Union[str, bytes]], file_type: Optional[str]) -> Union[str, List[Dict[str, Any]]]:
        """Process file attachment and return modified prompt."""
        if file is None:
            return prompt
            
        try:
            # Determine file type
            if file_type is None and isinstance(file, str):
                file_type = file.split(".")[-1].lower()
                
            if not file_type:
                raise ChatError("File type could not be determined")
                
            conf = Conf()
            
            # Handle text files
            if file_type in conf["supported_files"]["text"]:
                return self._process_text_file(prompt, file)
            # Handle image files
            elif file_type in conf["supported_files"]["image"]:
                return self._process_image_file(prompt, file)
            else:
                raise ChatError(f"Unsupported file type: {file_type}")
                
        except Exception as e:
            raise ChatError(f"Failed to process file attachment: {e}")

    def _process_text_file(self, prompt: str, file: Union[str, bytes]) -> str:
        """Process text file attachment."""
        try:
            if isinstance(file, str):
                # Handle as file path or URL
                if os.path.exists(file):
                    with open(file, "r", encoding="utf-8") as f:
                        file_content = f.read()
                else:
                    # Assume it's a URL or remote file
                    file_content = file
            else:
                # Handle as bytes
                file_content = file.decode("utf-8")
                
            return prompt + Conf()["default_prompt"]["file"] + file_content
            
        except Exception as e:
            raise ChatError(f"Failed to process text file: {e}")

    def _process_image_file(self, prompt: str, file: Union[str, bytes]) -> List[Dict[str, Any]]:
        """Process image file attachment."""
        try:
            if isinstance(file, str):
                # Handle as file path or URL
                if os.path.exists(file):
                    # Convert local file to base64
                    with open(file, "rb") as f:
                        file_bytes = f.read()
                    image_url = f"data:image/jpeg;base64,{base64.b64encode(file_bytes).decode('utf-8')}"
                else:
                    # Assume it's already a URL
                    image_url = file
            else:
                # Handle as bytes
                image_url = f"data:image/jpeg;base64,{base64.b64encode(file).decode('utf-8')}"
                
            return [
                {"type": "text", "text": prompt},
                {"type": "image_url", "image_url": image_url}
            ]
            
        except Exception as e:
            raise ChatError(f"Failed to process image file: {e}")

    def _prepare_messages(self, prompt: Union[str, List[Dict[str, Any]]]) -> List[Dict[str, Any]]:
        """Prepare messages for the API call."""
        try:
            messages = self._history.load(self.chat_id)
            messages = [{'role': d.get('role'), 'content': d.get('content')} for d in messages]

            messages.append({"role": "user", "content": prompt})
            print(messages)
            # Apply history summarization if enabled
            if self._history_summarizer is not None:
                return self._apply_history_summarization(messages)
            else:
                return messages
                
        except Exception as e:
            raise ChatError(f"Failed to prepare messages: {e}")

    def _apply_history_summarization(self, messages: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """Apply history summarization to messages."""
        try:
            summarized = self._history_summarizer.summarize(messages)
            ask_messages = [messages[0], messages[-1]]
            ask_messages[0]["content"] += Conf()["default_prompt"]["summary"] + summarized
            return ask_messages
        except Exception as e:
            logger.warning(f"History summarization failed, using full history: {e}")
            return messages

    def _store_messages(self, messages: List[Dict[str, Any]], response_content: str) -> None:
        """Store messages in history."""
        try:
            messages.append({"role": "assistant", "content": response_content})
            self._history.store(self.chat_id, messages[-2:])
        except Exception as e:
            logger.error(f"Failed to store messages: {e}")

    def ask(self, prompt: str, file: Optional[Union[str, bytes]] = None, file_type: Optional[str] = None, return_history: bool = False) -> Union[str, List[Dict[str, Any]]]:
        """
        Ask the model a question.

        Parameters
        ----------
        prompt : str
            The question to ask the model
        file : str | bytes, optional
            The file to attach to the message, if it's a string it will be treated as a local path or remote url to a file
        file_type : str, optional
            The type of the file
        return_history : bool, optional
            Whether to return the full history of the chat or only the response

        Returns
        -------
        str | List[Dict[str, Any]]
            The response from the model or the full history if return_history is True

        Raises
        ------
        ChatError
            If the request fails or parameters are invalid
        """
        try:
            # Process file attachment
            processed_prompt = self._process_file_attachment(prompt, file, file_type)
            
            # Prepare messages
            messages = self._prepare_messages(processed_prompt)
            
            # Make API call
            response = self._model.ask(messages)
            response_content = response["response"]
            
            # Store messages
            self._store_messages(messages, response_content)
            
            if return_history:
                return messages
            else:
                return response_content
                
        except Exception as e:
            logger.error(f"Ask request failed: {e}")
            raise ChatError(f"Ask request failed: {e}")

    async def ask_stream(self, prompt: str, file: Optional[Union[str, bytes]] = None, file_type: Optional[str] = None) -> AsyncGenerator[str, None]:
        """
        Ask the model a question and stream the response.

        Parameters
        ----------
        prompt : str
            The question to ask the model
        file : str | bytes, optional
            The file to attach to the message

        Yields
        ------
        str
            JSON string containing streaming response chunks

        Raises
        ------
        ChatError
            If the request fails or parameters are invalid
        """
        try:
            # Process file attachment
            processed_prompt = self._process_file_attachment(prompt, file, file_type)
            
            # Prepare messages
            messages = self._prepare_messages(processed_prompt)
            
            # Make streaming API call
            response = self._model.ask_stream(messages)

            response_text = ""
            async for chunk in response: 
                if "delta" in chunk:
                    response_text += chunk["delta"]
                yield chunk

            # Store messages
            self._store_messages(messages, response_text)
            
        except Exception as e:
            logger.error(f"Stream request failed: {e}")
            raise ChatError(f"Stream request failed: {e}")

    async def ask_async(self, prompt: str, file: Optional[Union[str, bytes]] = None, file_type: Optional[str] = None) -> AsyncGenerator[str, None]:
        """
        Ask the model a question and stream the response asynchronously.

        Parameters
        ----------
        prompt : str
            The question to ask the model
        file : str | bytes, optional
            The file to attach to the message
        file_type : str, optional
            The type of the file

        Yields
        ------
        str
            Response chunks as strings

        Raises
        ------
        ChatError
            If the request fails or parameters are invalid
        """
        try:
            # Process file attachment
            processed_prompt = self._process_file_attachment(prompt, file, file_type)
            
            # Prepare messages
            messages = self._prepare_messages(processed_prompt)
            
            # Make streaming API call
            response = await acompletion(
                model=self._model,
                messages=messages,
                stream=True,
                max_tokens=self._max_tokens
            )

            response_text = ""
            async for part in response:
                content = part["choices"][0]["delta"]["content"] or ""
                response_text += content
                yield content
                
            # Store messages
            self._store_messages(messages, response_text)
            
        except Exception as e:
            logger.error(f"Async request failed: {e}")
            raise ChatError(f"Async request failed: {e}")

    def get_chat_info(self) -> Dict[str, Any]:
        """
        Get information about the current chat.

        Returns
        -------
        Dict[str, Any]
            Dictionary containing chat information
        """
        return {
            "chat_id": self.chat_id,
            "provider": self._provider,
            "model": self._model_name,
            "max_tokens": self._max_tokens,
            "has_history_summarizer": self._history_summarizer is not None
        }

    def clear_history(self) -> None:
        """
        Clear the chat history.

        Raises
        ------
        ChatError
            If clearing history fails
        """
        try:
            self._history.clear(self.chat_id)
            logger.info(f"Cleared history for chat {self.chat_id}")
        except Exception as e:
            logger.error(f"Failed to clear history: {e}")
            raise ChatError(f"Failed to clear history: {e}")